import numpy as np
from scipy.special import betainc as betai
from sklearn.preprocessing import OneHotEncoder

from Orange.data import Domain, Table, ContinuousVariable, DiscreteVariable, Variable
from Orange.data.util import SharedComputeValue
from Orange.preprocess.preprocess import Preprocess, Continuize
from Orange.preprocess.score import Scorer

# Link / inverse link functions
LINK_IDENTITY = "Identity link"
LINK_LOG = "Log link"

LINKS = {
    LINK_IDENTITY: lambda x: x,
    LINK_LOG: np.log
}

INV_LINKS = {
    LINK_IDENTITY: lambda x: x,
    LINK_LOG: np.exp
}

# Percentage standard deviation
PERC_SD = 0.01


class ScBatchScorer(Scorer):
    """
    For Continuous batch variables,
        calculate a percentage of genes significantly (p < alpha)
        correlated with the variable.

    For Discrete batch variables,
        calculate the *relative size of union* of
        *significantly correlated genes* (p < alpha) for each value (one-hot encoding).
    """

    feature_type = Variable
    class_type = Variable
    supports_sparse_data = True
    friendly_name = "score"
    name = "ScBatchScore"

    @staticmethod
    def _std(A):
        """ Numpy implementation fails on matrices with one column. """
        return np.power(np.var(A, axis=0), 0.5)

    @staticmethod
    def correlations(A, B):
        """ Fast approximation to Pearson corr. and p-values for variables in two matrices."""
        # Correlation coefficients
        M = (A - A.mean(axis=0)).T.dot(B - B.mean(axis=0)) / A.shape[0]
        S = np.outer(ScBatchScorer._std(A), ScBatchScorer._std(B))
        S[S == 0] = 1.0
        M[S == 0] = 0.0
        rf = M / S

        # P-values
        df = A.shape[0] - 2
        ts = rf * rf * (df / (1 - rf * rf))
        pf = betai(0.5 * df, 0.5, np.array(df / (df + ts), dtype=float))
        return rf, pf

    def __init__(self, alpha=0.05):
        self.alpha = alpha

    def score_data(self, data, feature):
        if feature is None:
            raise ValueError("Scorer %s computes on a per-feature basis. ", self.__class__)
        if not all((isinstance(att, ContinuousVariable) for att in data.domain.attributes)):
            raise ValueError("All variables in the data must be Continuous!")

        a = data.get_column_view(feature.name)[0].reshape((len(data), 1))
        if isinstance(feature, ContinuousVariable):
            _, p = self.correlations(a, data.X)
            w = (p < self.alpha).mean()
        else:
            B = OneHotEncoder(sparse=False).fit_transform(a)
            _, P = self.correlations(B, data.X)
            w = (P < self.alpha).sum(axis=0).astype(bool).mean()
        return w

    def __call__(self, data, feature=None):
        return self.score_data(data, feature)



class ScBatchShared(SharedComputeValue):
    """Places the values of shared data within the corresponding variable column."""
    def compute(self, data, shared_data):
        assert self.variable is not None
        return shared_data.get_column_view(self.variable)[0] if self.variable in shared_data.domain else np.nan


class SCBatchNormalizer(Preprocess):
    """ Instantiate a new domain with transformations defined by the model. """
    def __init__(self, link=LINK_IDENTITY, nonzero_only=True, batch_vars=()):
        self.link = link
        self.nonzero_only = nonzero_only
        self.batch_vars = batch_vars

    def __call__(self, data):
        proj = ScBatchNormalizeModel(self.link, self.nonzero_only, self.batch_vars)
        proj.fit(data)
        attributes = [var.copy(compute_value=ScBatchShared(proj, variable=var))
                      for var in data.domain.attributes]
        for var in attributes:
            var.number_of_decimals = max(3, var.number_of_decimals)
        normalized_domain = Domain(
            attributes, data.domain.class_vars, data.domain.metas)
        return data.transform(normalized_domain)


class ScBatchNormalizeModel:

    def __init__(self, link=LINK_IDENTITY, nonzero_only=True, batch_vars=()):
        """
        :param link: Link function key.
        :param nonzero_only: Fit only on non-zero values.
        :param batch_vars: *names* of batch variables (must be meta).
        """
        if link == LINK_LOG and not nonzero_only:
            raise ValueError("Log link must be used with nonzero_only=True !")
        self.link = link
        self.nonzero_only = nonzero_only
        self.batch_vars = batch_vars
        self.models = dict()

    def _design_matrix(self, data):
        """ Create a design matrix with Continuized variables and a bias term.
            Can contain both meta and class variables. """
        assert len(self.batch_vars) > 0
        df = data[:, self.batch_vars]
        dom = df.domain.metas + df.domain.class_vars
        X = np.hstack((df.metas,
                       df.Y.reshape((len(data), len(df.domain.class_vars)))))
        Z = Continuize()(Table.from_numpy(domain=Domain(attributes=dom), X=X)).X
        return np.hstack((np.ones((len(df), 1)), Z))

    def fit(self, data):
        """ Fit one model per gene with least-squares. """
        atts = data.domain.attributes
        assert all([isinstance(a, ContinuousVariable) for a in atts])
        assert all([isinstance(data.domain[b], ContinuousVariable)
                    or isinstance(data.domain[b], DiscreteVariable) for b in self.batch_vars])
        if len(self.batch_vars) == 0:
            return
        Z = self._design_matrix(data)
        k = Z.shape[1]
        if self.nonzero_only:
            if np.any(data.X < 0):
                if self.link == LINK_IDENTITY:
                    raise ValueError("Data contains negative values. "
                                     "Remove 'nonzero_true' option.")
                else:
                    raise ValueError("Data contains negative values.")
            for i, a in enumerate(atts):
                nz = np.where(data.X[:, i])[0]
                Y = LINKS[self.link](data.X[:, i][nz])
                w = np.linalg.lstsq(Z[nz], Y, rcond=None)[0].reshape((k, 1)) \
                    if len(nz) else np.zeros((k, 1))
                self.models[a.name] = w
        else:
            Y = LINKS[self.link](data.X)
            if not (np.all(np.isfinite(Y)) and np.all(np.logical_not(np.isnan(Y)))):
                raise ValueError("Transformed data contains NaN/inf values. "
                                 "Use a different link function.")
            W = np.linalg.lstsq(Z, Y, rcond=None)[0]
            self.models = dict(((a.name, w.reshape((k, 1))) for a, w in zip(atts, W.T)))

    def transform(self, data):
        """ Apply transformation. Genes in new data must have had been available to fit.

        In case of linear regression (identity link), with nonzero_only=True,
        make sure zeros remains zeros and the original non-zeros are all larger than zero.

        This combination must not be used if the original data contains negative
        values.
        """
        if len(self.batch_vars) == 0:
            return data
        else:
            atts = data.domain.attributes
            assert all((a.name in self.models for a in atts))
            assert all(b in data.domain for b in self.batch_vars)
            Z = self._design_matrix(data)
            W = np.hstack((self.models[a.name].reshape((Z.shape[1], 1)) for a in atts))
            Xc = data.X.copy()
            if self.nonzero_only:
                nz = np.where(Xc)
                Xc[nz] = INV_LINKS[self.link](LINKS[self.link](Xc[nz]) - Z.dot(W)[nz])
                if self.link == LINK_IDENTITY:
                    mask = Xc != 0
                    sd = Xc.std() * PERC_SD
                    Xc = Xc - (Xc.min(axis=0) * mask) + (sd * mask)
            else:
                Xc = INV_LINKS[self.link](LINKS[self.link](Xc) - Z.dot(W))
            new_data = data.copy()
            new_data.X = Xc
            return new_data

    def __call__(self, data):
        """ Transform data (alias). """
        return self.transform(data)



import itertools as it
import numpy as np
from scipy.stats import pearsonr
from sklearn.decomposition import TruncatedSVD

import Orange.statistics.util as ut


def _standardize(X, tol=1e-8):
    """ Standardize matrix X avoiding NaNs and small values. """
    s = ut.std(X, axis=0)
    m = np.zeros(*s.shape)
    valid = s > tol
    m[np.logical_not(valid)] = 0
    m[valid] = 1.0 / s[valid]
    return (X - X.mean(axis=0)) * m


class SVDCCA:
    """ Two data set CCA approximated via SVD. Follows
        Butler, A. et al. (2018) "Integrating single-cell transcriptomic data across different conditions,
        technologies, and species",  Nature Biotechnology, (July 2017). doi: 10.1038/nbt.4096.
    """

    def __init__(self, n_components, random_state=None, standardize=True):
        self.n_components = n_components
        self.standardize = standardize
        self.random_state = random_state
        self.correlations = None

    def fit_transform(self, X, Y):
        if self.standardize:
            X = _standardize(X)
            Y = _standardize(Y)
        K = X.dot(Y.T)
        model = TruncatedSVD(n_components=self.n_components,
                             random_state=self.random_state)
        U = model.fit_transform(K)
        U = U / np.linalg.norm(U, axis=0)
        V = model.components_.T
        self.correlations = np.array([pearsonr(u.dot(X), v.dot(Y))[0]
                                      for u, v in zip(U.T, V.T)])
        return U, V


class MultiCCA:

    """ Approximation to penalized CCA for multiple datasets.
        Follows Seurat MultiCCA method.
    """

    def __init__(self, n_components, max_iter=25, tol=1e-3, random_state=None, standardize=True):
        self.n_components = n_components
        self.max_iter = max_iter
        self.tol = tol
        self.random_state = random_state
        self.standardize = standardize
        self.correlations = None

    @staticmethod
    def _objective(Xs, ws):
        """ CCA objective function. """
        c = 0
        for i, j in it.combinations(range(len(Xs)), 2):
            c += ws[i].T.dot(Xs[i]).dot(Xs[j].T.dot(ws[j]))
        return c

    def fit_transform(self, Xs):
        """
        Optimize each CC components and return per-data set projections.
        :param Xs: List of matrices with the same number of columns.
        :return: CCA subspace.
        """
        p = len(Xs)
        Ws = [np.zeros((X.shape[0], self.n_components)) for X in Xs]
        if self.standardize:
            Xs = list(map(_standardize, Xs))
        Ws_init = [TruncatedSVD(n_components=self.n_components,
                                random_state=self.random_state).fit_transform(X) for X in Xs]
        correlations = np.zeros((self.n_components,))


        # Optimize each CC component individually
        for cc in range(self.n_components):
            w_cur = [Wi[:, cc] / np.linalg.norm(Wi[:, cc]) for Wi in Ws_init]
            for itr in range(self.max_iter):
                o1 = self._objective(Xs, w_cur)
                for i in range(p):
                    wi = 0
                    for j in range(p):
                        if i == j:
                            continue
                        wj = w_cur[j]
                        Dj = np.diag(np.diagonal(Ws[i].T.dot(Xs[i]).dot(Xs[j].T.dot(Ws[j]))))
                        wi += Xs[i].dot((Xs[j].T.dot(wj))) - Ws[i].dot(Dj).dot(Ws[j].T).dot(wj)
                    w_cur[i] = wi / np.linalg.norm(wi)
                o2 = self._objective(Xs, w_cur)
                if abs(o2-o1) / abs(o1) < self.tol:
                    break
            for i in range(p):
                Ws[i][:, cc] = w_cur[i]

            # Compute average correlations
            n_pairs = p * (p - 1) / 2
            for i, j in it.combinations(range(p), 2):
                wi = Ws[i][:, cc].T.dot(Xs[i])
                wj = Ws[j][:, cc].T.dot(Xs[j])
                correlations[cc] += pearsonr(wi, wj)[0] / n_pairs

        # Orientate vectors
        s = np.sign(Ws[0][0, :])
        for i in range(p):
            Ws[i] = Ws[i] * s

        self.correlations = correlations
        return Ws



import numpy as np

from Orange.data import Table
from Orange.widgets.tests.base import WidgetTest

from orangecontrib.single_cell.preprocess.scpreprocess import (
    LogarithmicScale, Binarize, Normalize, DropoutGeneSelection,
    NormalizeGroups, NormalizeSamples, Standardize, SelectMostVariableGenes)
from orangecontrib.single_cell.widgets import owscpreprocess
from orangecontrib.single_cell.widgets.owscpreprocess import OWscPreprocess


class TestOWscPreprocess(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(OWscPreprocess)

    def test_available_preprocessors(self):
        self.assertEqual(self.widget.preprocessors.rowCount(), 6)

    def test_default_preprocessors(self):
        self.assertEqual(self.widget.preprocessormodel.rowCount(), 4)

    def test_preprocess(self):
        data = Table("iris")
        self.send_signal(self.widget.Inputs.data, data)
        pp_data = self.get_output(self.widget.Outputs.preprocessed_data)
        self.assertIsInstance(pp_data, Table)
        self.assertNotEqual(pp_data, data)

    def test_discrete_attributes(self):
        self.send_signal(self.widget.Inputs.data, Table("iris"))
        pp_data = self.get_output(self.widget.Outputs.preprocessed_data)
        self.assertIsNotNone(pp_data)
        self.send_signal("Data", Table("heart_disease"))
        pp_data = self.get_output(self.widget.Outputs.preprocessed_data)
        self.assertIsNone(pp_data)
        self.assertTrue(self.widget.Error.discrete_attributes.is_shown())
        self.send_signal("Data", None)
        self.assertFalse(self.widget.Error.discrete_attributes.is_shown())

    def test_missing_values(self):
        data = Table("iris")
        data[0, 3] = np.nan
        self.send_signal(self.widget.Inputs.data, data)
        pp_data = self.get_output(self.widget.Outputs.preprocessed_data)
        self.assertIsInstance(pp_data, Table)
        self.assertNotEqual(pp_data, data)
        self.assertTrue(self.widget.Warning.missing_values.is_shown())
        self.send_signal("Data", None)
        self.assertFalse(self.widget.Warning.missing_values.is_shown())
        pp_data = self.get_output(self.widget.Outputs.preprocessed_data)
        self.assertIsNone(pp_data)

    def test_dropout_warning(self):
        pp_setting = {"preprocessors": [('preprocess.dropout', {})]}
        settings = {"storedsettings": pp_setting}
        w = self.create_widget(OWscPreprocess, stored_settings=settings)
        self.send_signal(w.Inputs.data, Table("iris"), widget=w)
        pp_data = self.get_output(w.Outputs.preprocessed_data, widget=w)
        self.assertIsInstance(pp_data, Table)
        self.assertTrue(w.Warning.dropout_warning.is_shown())
        self.send_signal(w.Inputs.data, None, widget=w)
        self.assertFalse(w.Warning.dropout_warning.is_shown())
        pp_data = self.get_output(w.Outputs.preprocessed_data, widget=w)
        self.assertIsNone(pp_data, Table)

    def test_bad_pp_combination(self):
        pp_setting = {"preprocessors": [("preprocess.normalize", {}),
                                        ('preprocess.dropout', {})]}
        settings = {"storedsettings": pp_setting}
        w = self.create_widget(OWscPreprocess, stored_settings=settings)
        self.send_signal(w.Inputs.data, Table("iris"), widget=w)
        self.assertTrue(w.Warning.bad_pp_combination.is_shown())

        pp_setting = {"preprocessors": [('preprocess.dropout', {})]}
        settings = {"storedsettings": pp_setting}
        w = self.create_widget(OWscPreprocess, stored_settings=settings)
        self.send_signal(w.Inputs.data, Table("iris"), widget=w)
        self.assertFalse(w.Warning.bad_pp_combination.is_shown())

        pp_setting = {"preprocessors": [("preprocess.dropout", {}),
                                        ('preprocess.normalize', {})]}
        settings = {"storedsettings": pp_setting}
        w = self.create_widget(OWscPreprocess, stored_settings=settings)
        self.send_signal(w.Inputs.data, Table("iris"), widget=w)
        self.assertFalse(w.Warning.bad_pp_combination.is_shown())


class TestLogarithmicScaleEditor(WidgetTest):
    def test_editor_default(self):
        widget = owscpreprocess.LogarithmicScaleEditor()
        p = widget.createinstance(widget.parameters())
        self.assertIsInstance(p, LogarithmicScale)
        self.assertEqual(p.base, LogarithmicScale.BinaryLog)


class TestBinarizeEditor(WidgetTest):
    def test_editor_default(self):
        widget = owscpreprocess.BinarizeEditor()
        p = widget.createinstance(widget.parameters())
        self.assertIsInstance(p, Binarize)
        self.assertEqual(p.condition, Binarize.GreaterOrEqual)
        self.assertEqual(p.threshold, 1)


class TestNormalizeEditor(WidgetTest):
    def test_editor_default(self):
        pp = owscpreprocess.OWscPreprocess()
        widget = owscpreprocess.NormalizeEditor(master=pp)
        p = widget.createinstance(widget.parameters())
        self.assertIsInstance(p, NormalizeSamples)
        self.assertEqual(p.method, Normalize.CPM)

    def test_editor_normalize_groups(self):
        data = Table("iris")
        pp = owscpreprocess.OWscPreprocess()
        pp.set_data(data)
        widget = owscpreprocess.NormalizeEditor(master=pp)
        params = {"method": Normalize.Median,
                  "group_var": data.domain.class_var,
                  "group_by": True}
        widget.setParameters(params)
        p = widget.createinstance(widget.parameters())
        self.assertIsInstance(p, NormalizeGroups)
        self.assertEqual(p.group_var, data.domain.class_var)
        self.assertEqual(p.method, Normalize.Median)


class TestStandardizeEditor(WidgetTest):
    def test_editor_default(self):
        widget = owscpreprocess.StandardizeEditor()
        p = widget.createinstance(widget.parameters())
        self.assertIsInstance(p, Standardize)
        self.assertIsNone(p.lower_bound)
        self.assertIsNone(p.upper_bound)


class TestSelectGenesEditor(WidgetTest):
    def test_editor_default(self):
        widget = owscpreprocess.SelectGenesEditor()
        p = widget.createinstance(widget.parameters())
        self.assertIsInstance(p, SelectMostVariableGenes)
        self.assertEqual(p.method, SelectMostVariableGenes.Dispersion)
        self.assertEqual(p.n_genes, 1000)
        self.assertEqual(p.n_groups, 20)


class TestDropoutEditor(WidgetTest):
    def test_editor_default(self):
        widget = owscpreprocess.DropoutEditor()
        p = widget.createinstance(widget.parameters())
        self.assertIsInstance(p, DropoutGeneSelection)
        self.assertEqual(p.n_genes, 1000)



import os
import random
from itertools import chain
from typing import Dict, Tuple

import numpy as np
import pandas as pd
import scipy.io
import scipy.sparse as sp
import loompy as lp
import xlrd

from Orange.data import (
    ContinuousVariable, DiscreteVariable, Domain, Table
)
from Orange.data.io import (
    Compression, open_compressed, PickleReader,
    guess_data_type, sanitize_variable
)
from Orange.widgets.utils.filedialogs import RecentPath


def separator_from_filename(file_name):
    """Get separator from file extension

    :param file_name: str
    :return: str
    """
    return "," if os.path.splitext(file_name)[1] == ".csv" else "\t"


def get_data_loader(file_name):
    """Get instance of data loader according to file extension

    :param file_name: str
    :return: Loader
    """
    base, ext = os.path.splitext(file_name)
    if ext in Compression.all:
        _, ext = os.path.splitext(base)
    if ext == ".mtx":
        return MtxLoader(file_name)
    elif ext == ".count":
        return CountLoader(file_name)
    elif ext == ".csv":
        return CsvLoader(file_name)
    elif ext in (".pkl", ".pickle"):
        return PickleLoader(file_name)
    elif ext in (".xls", ".xlsx"):
        return ExcelLoader(file_name)
    elif ext == ".loom":
        return LoomLoader(file_name)
    elif ext == ".h5ad":
        return H5ADLoader(file_name)
    else:
        return Loader(file_name)


class Loader:
    """Class loads sample file (and it's supplementary annotation
    files, if present) and creates Orange.data.Table out of its data
    """
    separator = "\t"

    def __init__(self, file_name=""):
        # file parameters
        self._file_name = file_name
        self._recent_path = RecentPath.create(file_name, [])
        self.file_size = None
        self.n_rows = None
        self.n_cols = None
        self.sparsity = None
        self._set_file_size()
        self._set_file_parameters()

        # reading parameters
        self._leading_cols = 0
        self._leading_rows = 0
        self._use_rows_mask = None
        self._use_cols_mask = None
        self._row_annot_header = 0
        self._row_annot_columns = None
        self._col_annot_header = 0
        self._col_annot_columns = None

        # GUI parameters
        self.FIXED_FORMAT = True
        self.ENABLE_ANNOTATIONS = True
        self.header_rows_count = None
        self.header_cols_count = None
        self.transposed = None
        self.sample_rows_enabled = None
        self.sample_cols_enabled = None
        self.sample_rows_p = None
        self.sample_cols_p = None
        self.row_annotations_enabled = True
        self.row_annotation_file = None  # type: RecentPath
        self.col_annotations_enabled = True
        self.col_annotation_file = None  # type: RecentPath

        # errors
        self.errors = {}  # type: Dict[str, Tuple]
        self.__reset_error_messages()

    @property
    def recent_path(self):
        return self._recent_path

    @recent_path.setter
    def recent_path(self, path):
        self._recent_path = path
        self._file_name = path.abspath

    @property
    def n_genes(self):
        return self.n_rows if self.transposed else self.n_cols

    @property
    def n_cells(self):
        return self.n_cols if self.transposed else self.n_rows

    @property
    def leading_rows(self):
        return self._leading_rows

    @leading_rows.setter
    def leading_rows(self, value):
        self._leading_rows = value

    @property
    def leading_cols(self):
        return self._leading_cols

    @leading_cols.setter
    def leading_cols(self, value):
        self._leading_cols = value

    def _set_file_size(self):
        try:
            st = os.stat(self._file_name)
        except OSError:
            pass
        else:
            self.file_size = st.st_size

    def _set_file_parameters(self):
        try:
            args = (self._file_name,)
            col_kwargs = {"sep": self.separator, "index_col": None, "nrows": 1}
            row_kwargs = {"sep": self.separator, "usecols": [1]}
            self.n_cols = self.df_read_func(*args, **col_kwargs).shape[1]
            self.n_rows = self.df_read_func(*args, **row_kwargs).shape[0]
            self._set_sparsity()
        except Exception:
            pass

    def _set_sparsity(self):
        """Get approximate sparsity if number of columns is bigger than 100."""
        if self.n_cols is not None and self.n_rows is not None:
            n_max = 100
            np.random.seed(42)
            use_cols = np.arange(1, self.n_cols - 1) if self.n_cols < n_max \
                else np.random.randint(1, self.n_cols - 1, n_max)

            data = self.df_read_func(
                self._file_name, usecols=use_cols, sep=self.separator,
                skipfooter=(max(0, self.n_rows - n_max))
            )

            n_all = data.shape[0] * data.shape[1]
            self.sparsity = (n_all - np.count_nonzero(data)) / n_all

    def _load_data(self, skip_row=None, header_rows=None, header_cols=None,
                   use_cols=None, **kwargs):
        df = self.df_read_func(
            self._file_name, sep=self.separator, index_col=header_cols,
            header=header_rows, skiprows=skip_row, usecols=use_cols
        )
        if skip_row is not None:
            self._use_rows_mask = np.array(self._use_rows_mask, dtype=bool)
        if self.transposed:
            df = df.transpose()
            self._use_rows_mask, self._use_cols_mask = \
                self._use_cols_mask, self._use_rows_mask
            self.leading_rows, self.leading_cols = \
                self.leading_cols, self.leading_rows

        attrs = [ContinuousVariable.make(str(g)) for g in df.columns]

        return attrs, df.values, df.iloc[:, :0], df.index

    @staticmethod
    def df_read_func(*args, **kwargs):
        return pd.read_csv(*args, **kwargs)

    def __call__(self):
        self.__reset_error_messages()

        rst = np.random.RandomState(0x667)
        if self.transposed:
            skip_row, skip_col = self.__skip_col(rst), self.__skip_row(rst)
        else:
            skip_col, skip_row = self.__skip_col(rst), self.__skip_row(rst)

        header_rows = self.__header_rows()
        header_cols, header_cols_indices = self.__header_cols()

        usecols, skip_col, skip_row, header_cols, header_rows = \
            self.__update_reading_parameters(
                skip_col, skip_row, header_cols_indices,
                header_cols, header_rows
            )

        try:
            attrs, X, meta_df, meta_df_index = self._load_data(
                skip_row=skip_row, skip_col=skip_col,
                header_rows=header_rows, header_cols=header_cols,
                use_cols=usecols, transpose=self.transposed
            )
        except Exception as e:
            self.errors["reading_error"] = (e, None)
            return None

        meta_parts = (meta_df,)
        if self.row_annotations_enabled and self.row_annotation_file:
            meta_df, row_annot_df = self.__update_metas(
                meta_df, meta_df_index, X)
            if row_annot_df is not None:
                meta_parts = (meta_df, row_annot_df)

        if self.col_annotations_enabled and self.col_annotation_file:
            attrs = self.__update_attributes(attrs, X)

        return self.__into_orange_table(attrs, X, meta_parts)

    def __header_rows(self):
        header_rows = self.header_rows_count
        header_rows_indices = []
        if header_rows == 0:
            header_rows = None
        elif header_rows == 1:
            header_rows = 0
            header_rows_indices = [0]
        else:
            header_rows = list(range(header_rows))
            header_rows_indices = header_rows
        self.leading_rows = len(header_rows_indices)
        return header_rows

    def __header_cols(self):
        header_cols = self.header_cols_count
        header_cols_indices = []
        if header_cols == 0:
            header_cols = None
        elif header_cols == 1:
            header_cols = 0
            header_cols_indices = [0]
        else:
            header_cols = list(range(header_cols))
            header_cols_indices = header_cols
        self.leading_cols = len(header_cols_indices)
        return header_cols, header_cols_indices

    def __skip_row(self, rstate):
        skip_row = None
        if self.sample_rows_enabled:
            p = self.sample_rows_p
            if p < 100:
                def skip_row(i, p=p):
                    return i > 3 and rstate.uniform(0, 100) > p
        return skip_row

    def __skip_col(self, rstate):
        skip_col = None
        if self.sample_cols_enabled:
            p = self.sample_cols_p
            if p < 100:
                def skip_col(i, p=p):
                    return i > 3 and rstate.uniform(0, 100) > p
        return skip_col

    def __update_reading_parameters(self, skip_col, skip_row, header_indices,
                                    header_cols, header_rows):
        self._use_rows_mask = None
        self._use_cols_mask = None
        usecols = None

        if skip_col is not None:
            self._use_cols_mask = np.array([
                not skip_col(i) or i in header_indices
                for i in range(self.n_cols)], dtype=bool)
            usecols = np.flatnonzero(self._use_cols_mask)

        if skip_row is not None:
            self._use_rows_mask = []  # record the used rows

            def skip_row(i, test=skip_row):
                r = test(i)
                self._use_rows_mask.append(r)
                return r

        # if skipping more than one leading columns/rows, use the
        # first one when creating domain
        if isinstance(header_cols, list) and self.transposed:
            if skip_col is None:
                mask = np.array([i not in header_cols[1:] for i
                                 in range(self.n_cols)], dtype=bool)
            else:
                mask = np.array([i not in header_cols[1:] and not skip_col(i)
                                 for i in range(self.n_cols)], dtype=bool)
            self._use_cols_mask = mask
            usecols = np.flatnonzero(mask)
            header_cols = 0

        if isinstance(header_rows, list) and not self.transposed:
            if skip_row is None:
                skip_row = header_rows[1:]
            else:
                self._use_rows_mask = []

                def skip_row(i, test=skip_row, header=header_rows[1:]):
                    r = test(i) or i in header
                    self._use_rows_mask.append(r)
                    return r

            header_rows = 0

        return usecols, skip_col, skip_row, header_cols, header_rows

    def __update_metas(self, meta_df, meta_df_index, X):
        file_name = self.row_annotation_file.abspath
        row_annot_df = pd.read_csv(
            file_name, sep=separator_from_filename(file_name),
            header=self._row_annot_header, names=self._row_annot_columns
        )
        if self._use_rows_mask is not None:
            # NOTE: we account for column header/ row index
            expected = len(self._use_rows_mask) - self.leading_rows
        else:
            expected = X.shape[0]
        if len(row_annot_df) != expected:
            self.errors["row_annot_mismatch"] = (expected, len(row_annot_df))
            row_annot_df = None

        if row_annot_df is not None and self._use_rows_mask is not None:
            # use the same sample indices
            indices = np.flatnonzero(self._use_rows_mask[self.leading_rows:])
            row_annot_df = row_annot_df.iloc[indices]
            # if path.endswith(".count") and row_annot.endswith('.meta'):
            #     assert np.all(row_annot_df.iloc[:, 0] == df.index)

        if row_annot_df is not None and meta_df_index is not None:
            # Try to match the leading columns with the meta_df_index.
            # If found then drop the columns (or index if the level does
            # not have a name but the annotation col does)
            drop_cols = []
            drop_index_level = []
            for i in range(meta_df_index.nlevels):
                meta_df_level = meta_df_index.get_level_values(i)
                if np.all(row_annot_df.iloc[:, i] == meta_df_level):
                    if meta_df_level.name is None:
                        drop_index_level.append(i)
                    elif meta_df_level.name == row_annot_df.columns[i].name:
                        drop_cols.append(i)

            if drop_cols:
                row_annot_df = row_annot_df.drop(columns=drop_cols)

            if drop_index_level:
                for i in reversed(drop_index_level):
                    if isinstance(meta_df.index, pd.MultiIndex):
                        meta_df_index = meta_df_index.droplevel(i)
                    else:
                        assert i == 0
                        meta_df_index = pd.RangeIndex(meta_df_index.size)
                meta_df = pd.DataFrame({}, index=meta_df_index)
        return meta_df, row_annot_df

    def __update_attributes(self, attrs, X):
        file_name = self.col_annotation_file.abspath
        col_annot_df = pd.read_csv(
            file_name, sep=separator_from_filename(file_name),
            header=self._col_annot_header, names=self._col_annot_columns
        )
        if self._use_cols_mask is not None:
            expected = len(self._use_cols_mask) - self.leading_cols
        else:
            expected = X.shape[1]
        if len(col_annot_df) != expected:
            self.errors["col_annot_mismatch"] = (expected, len(col_annot_df))
            col_annot_df = None
        if col_annot_df is not None and self._use_cols_mask is not None:
            indices = np.flatnonzero(self._use_cols_mask[self.leading_cols:])
            col_annot_df = col_annot_df.iloc[indices]

        if col_annot_df is not None:
            assert len(col_annot_df) == X.shape[1]
            if not attrs and X.shape[1]:  # No column names yet
                attrs = [ContinuousVariable.make(str(v))
                         for v in col_annot_df.iloc[:, 0]]
            names = [str(c) for c in col_annot_df.columns]
            for var, values in zip(attrs, col_annot_df.values):
                var.attributes.update(
                    {n: v for n, v in zip(names, values)})
        return attrs

    def __into_orange_table(self, attrs, X, meta_parts):
        if not attrs and X.shape[1]:
            attrs = Domain.from_numpy(X).attributes

        try:
            metas = None
            M = None
            if meta_parts:
                meta_parts = [df_.reset_index() if not df_.index.is_integer()
                              else df_ for df_ in meta_parts]
                metas, M = self.__guess_metas(meta_parts)

            domain = Domain(attrs, metas=metas)
            table = Table.from_numpy(domain, X, None, M)
        except ValueError:
            table = None
            rows = self.leading_cols if self.transposed else self.leading_rows
            cols = self.leading_rows if self.transposed else self.leading_cols
            self.errors["inadequate_headers"] = (rows, cols)
        return table

    @staticmethod
    def __guess_metas(meta_parts):
        def guessed_var(i, var_name, dtype):
            if pd.core.dtypes.common.is_numeric_dtype(dtype):
                return ContinuousVariable.make(var_name)
            orig_values = M[:, i]
            val_map, values, var_type = guess_data_type(orig_values)
            values, variable = sanitize_variable(
                val_map, values, orig_values, var_type, {}, name=var_name)
            M[:, i] = values
            return variable

        M = np.hstack(tuple(df_.values for df_ in meta_parts))
        return [guessed_var(i, name, dtype) for i, (name, dtype) in
                enumerate(zip(chain(*(_.columns for _ in meta_parts)),
                              chain(*(_.dtypes for _ in meta_parts))))], M

    def __reset_error_messages(self):
        self.errors = {"row_annot_mismatch": (),
                       "col_annot_mismatch": (),
                       "inadequate_headers": (),
                       "reading_error": ()}

    def copy(self):
        loader = self.__class__(self._file_name)
        for key in vars(loader):
            setattr(loader, key, getattr(self, key))
        return loader


class MtxLoader(Loader):
    def __init__(self, file_name):
        super().__init__(file_name)
        self.header_rows_count = 0
        self.header_cols_count = 0
        self.FIXED_FORMAT = False
        self.transposed = True
        self._row_annot_header = None
        self._row_annot_columns = ["Barcodes"]
        self._col_annot_header = None
        self._col_annot_columns = ["Id", "Gene"]
        self._set_annotation_files()
        self._set_enable_annotations()

    @property
    def leading_rows(self):
        return self._leading_rows

    @leading_rows.setter
    def leading_rows(self, value):
        self._leading_rows = 0

    @property
    def leading_cols(self):
        return self._leading_cols

    @leading_cols.setter
    def leading_cols(self, value):
        self._leading_cols = 0

    def _set_annotation_files(self):
        dir_name, _ = os.path.split(self._file_name)
        genes_path = os.path.join(dir_name, "genes.tsv")
        if os.path.isfile(genes_path):
            self.col_annotation_file = RecentPath.create(genes_path, [])
        barcodes_path = os.path.join(dir_name, "barcodes.tsv")
        if os.path.isfile(barcodes_path):
            self.row_annotation_file = RecentPath.create(barcodes_path, [])

    def _set_enable_annotations(self):
        # 10x gene-barcode matrix
        # TODO: The genes/barcodes files should be unconditionally loaded
        # alongside the mtx. The row/col annotations might be used to
        # specify additional sources. For the time being they are put in
        # the corresponding comboboxes and made uneditable.
        if self.row_annotation_file is not None \
                and self.col_annotation_file is not None \
                and os.path.basename(
                self.col_annotation_file.abspath) == "genes.tsv" \
                and os.path.basename(
                self.row_annotation_file.abspath) == "barcodes.tsv":
            self.ENABLE_ANNOTATIONS = False

    def _set_file_parameters(self):
        try:
            with open_compressed(self._file_name, "rb") as f:
                self.n_rows, self.n_cols, non_zero_el = scipy.io.mminfo(f)[:3]
                all_el = self.n_rows * self.n_cols
                self.sparsity = (all_el - non_zero_el) / all_el
        except OSError:
            pass
        except ValueError:
            pass

    def _load_data(self, skip_row=None, skip_col=None, **kwargs):
        X = scipy.io.mmread(self._file_name)
        if self.transposed:
            X = X.T
        if skip_row is not None:
            self._use_rows_mask = np.array(
                [not skip_row(i) for i in range(X.shape[0])]
            )
            X = X.tocsr()[np.flatnonzero(self._use_rows_mask)]
        if skip_col is not None:
            self._use_cols_mask = np.array(
                [not skip_col(i) for i in range(X.shape[1])]
            )
            X = X.tocsc()[:, np.flatnonzero(self._use_cols_mask)]
        X = X.todense(order="F")
        if self._use_rows_mask is not None:
            meta_df = pd.DataFrame(
                {}, index=np.flatnonzero(self._use_rows_mask))
        else:
            meta_df = pd.DataFrame({}, index=pd.RangeIndex(X.shape[0]))
        return [], X, meta_df, meta_df.index


class CountLoader(Loader):
    def __init__(self, file_name):
        super().__init__(file_name)
        self.header_rows_count = 1
        self.header_cols_count = 1
        self.FIXED_FORMAT = False
        self.transposed = True
        self._set_annotation_files()

    def _set_annotation_files(self):
        dir_name, basename = os.path.split(self._file_name)
        basename_no_ext, _ = os.path.splitext(basename)
        meta_path = os.path.join(dir_name, basename_no_ext + ".meta")
        if os.path.isfile(meta_path):
            self.row_annotation_file = RecentPath.create(meta_path, [])


class CsvLoader(Loader):
    separator = ","


class PickleLoader(Loader):
    separator = None

    def __init__(self, file_name):
        super().__init__(file_name)
        self.header_rows_count = 0
        self.header_cols_count = 0
        self.FIXED_FORMAT = False
        self.ENABLE_ANNOTATIONS = False
        self.transposed = False
        self.row_annotations_enabled = False
        self.col_annotations_enabled = False

    def _set_file_parameters(self):
        pass

    def _load_data(self):
        random.seed(0)
        reader = PickleReader(self._file_name)
        table = reader.read()
        self.n_rows, self.n_cols = table.X.shape
        attrs = self.__attributes(table.domain.attributes)
        metas = table.domain.metas + table.domain.class_vars
        domain = Domain(attrs, metas=metas)
        return table.transform(domain)[self.__row_indices()]

    def __call__(self):
        return self._load_data()

    def __row_indices(self):
        indices = range(self.n_rows)
        if self.sample_rows_enabled:
            p = self.sample_rows_p
            if p < 100 and self.n_rows > 3:
                return random.sample(indices, int(self.n_rows * p / 100))
        return indices

    def __attributes(self, attributes):
        if self.sample_cols_enabled:
            p = self.sample_cols_p
            if p < 100 and self.n_cols > 3:
                return random.sample(attributes, int(self.n_cols * p / 100))
        return attributes


class ExcelLoader(Loader):
    def __init__(self, file_name):
        super().__init__(file_name)
        self.header_rows_count = 1
        self.header_cols_count = 1

    def _set_file_parameters(self):
        try:
            sheet = xlrd.open_workbook(self._file_name).sheet_by_index(0)
            self.n_cols = sheet.ncols
            self.n_rows = sheet.nrows
            self._set_sparsity()
        except Exception:
            pass

    @staticmethod
    def df_read_func(*args, **kwargs):
        kwargs.pop("sep")
        return pd.read_excel(*args, **kwargs)


class LoomLoader(Loader):
    def __init__(self, file_name):
        super().__init__(file_name)
        self.header_rows_count = 0
        self.header_cols_count = 0
        self.FIXED_FORMAT = False
        self.ENABLE_ANNOTATIONS = False
        self.transposed = True
        self.row_annotations_enabled = False
        self.col_annotations_enabled = False

    def _set_file_parameters(self):
        try:
            with lp.connect(self._file_name) as ds:
                self.n_rows, self.n_cols = ds.shape
                all_el = self.n_rows * self.n_cols
                self.sparsity = (all_el - ds.sparse().count_nonzero()) / all_el
        except OSError:
            pass

    def _load_data(self, skip_row=None, skip_col=None, **kwargs):
        with lp.connect(self._file_name) as ds:
            X = ds[:, :].T
            if skip_row is not None:
                mask = np.array([not skip_row(i) for i in range(X.shape[1])])
                self._use_rows_mask = mask
            else:
                self._use_rows_mask = np.ones(X.shape[1], dtype=bool)
            if skip_col is not None:
                mask = np.array([not skip_col(i) for i in range(X.shape[0])])
                self._use_cols_mask = mask
            else:
                self._use_cols_mask = np.ones(X.shape[0], dtype=bool)

            X = X[self._use_cols_mask, :]
            X = X[:, self._use_rows_mask]
            gene_names = ds.ra.Gene[self._use_rows_mask] \
                if hasattr(ds.ra, "Gene") else []
            attrs = [ContinuousVariable.make(str(g)) for g in gene_names]
            meta_df = pd.DataFrame({key: ds.ca[key][self._use_cols_mask]
                                    for key in ds.ca.keys()})

        return attrs, X, meta_df, meta_df.index


class H5ADLoader(Loader):
    def __init__(self, file_name):
        super().__init__(file_name)
        self.header_rows_count = 0
        self.header_cols_count = 0
        self.FIXED_FORMAT = False
        self.ENABLE_ANNOTATIONS = False
        self.transposed = False
        self.row_annotations_enabled = False
        self.col_annotations_enabled = False

    def _set_file_parameters(self):
        import anndata
        try:
            adata = anndata.read_h5ad(self._file_name)
            self.n_rows, self.n_cols = adata.shape
            all_el = self.n_rows * self.n_cols
            if sp.issparse(adata.X):
                self.sparsity = (all_el - adata.X.tocsr().count_nonzero()) / all_el
            else:
                self.sparsity = (all_el - np.count_nonzero(adata.X)) / all_el
        except OSError:
            pass

    def _load_data(self, skip_row=None, skip_col=None, **kwargs):
        import anndata
        adata = anndata.read_h5ad(self._file_name)

        if skip_row is not None:
            mask = np.array([not skip_row(i) for i in range(adata.shape[0])])
            self._use_rows_mask = mask
        else:
            self._use_rows_mask = np.ones(adata.shape[0], dtype=bool)
        if skip_col is not None:
            mask = np.array([not skip_col(i) for i in range(adata.shape[1])])
            self._use_cols_mask = mask
        else:
            self._use_cols_mask = np.ones(adata.shape[1], dtype=bool)

        adata = adata[self._use_rows_mask, :]
        adata = adata[:, self._use_cols_mask]

        attrs = [ContinuousVariable.make(str(g)) for g in adata.var_names]
        meta_df = pd.DataFrame({key: adata.obs[key] for key in adata.obs.keys()})

        X = adata.X.toarray() if sp.issparse(adata.X) else adata.X

        return attrs, X, meta_df, meta_df.index


class Concatenate:
    INTERSECTION, UNION = range(2)

    @classmethod
    def concatenate(cls, concat_type, data_collection):
        if not data_collection:
            return None

        concat_data, source_var = cls.append_source_name(*data_collection[0])
        for data, source_name in data_collection[1:]:
            attrs1 = concat_data.domain.attributes
            attrs2 = data.domain.attributes
            if concat_type == cls.INTERSECTION:
                attrs = set(attrs1).intersection(attrs2)
            elif concat_type == cls.UNION:
                attrs = set(attrs1 + attrs2)
            metas = set(concat_data.domain.metas + data.domain.metas)

            def key(var):
                return var.name if isinstance(var.name, str) else ""

            domain = Domain(sorted(attrs, key=key),
                            metas=sorted(metas, key=key))
            concat_data_t = concat_data.transform(domain)
            data_t = data.transform(domain)
            source_var.values + (source_name, )
            data_t[:, source_var] = np.full(
                (len(data), 1), len(source_var.values) - 1, dtype=object
            )
            concat_data = Table.concatenate((concat_data_t, data_t), axis=0)
        return concat_data

    @staticmethod
    def append_source_name(data, name):
        source_var = DiscreteVariable("source", values=[name])
        metas = data.domain.metas + (source_var,)
        domain = Domain(data.domain.attributes, metas=metas)
        data = data.transform(domain)
        data[:, source_var] = np.full((len(data), 1), 0, dtype=object)
        return data, source_var



import unittest
import numpy as np

from orangecontrib.single_cell.widgets.owscorecells import OWScoreCells
from Orange.widgets.tests.base import WidgetTest
from Orange.data import DiscreteVariable, ContinuousVariable, StringVariable, Domain, Table

from orangecontrib.bioinformatics.widgets.utils.data import (
    GENE_ID_ATTRIBUTE,
    GENE_ID_COLUMN,
    GENE_AS_ATTRIBUTE_NAME,
    TAX_ID,
)


class TestOWScoreCells(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(OWScoreCells)

        class_var = DiscreteVariable('Stage name', values=['STG1', 'STG2'])
        attributes = [ContinuousVariable('GeneName' + str(i)) for i in range(5)]

        for idx, attr in enumerate(attributes):
            attr.attributes['Entrez ID'] = idx

        domain = Domain(attributes, class_vars=class_var)

        self.data = Table.from_numpy(
            domain,
            [[1, 2, 3, 4, 5], [4, 4, 4, 4, 4], [2, 3, 1, 1, 1], [-1, 0, 1, 0, 0]],
            [[0], [0], [0], [1]]
        )
        self.data.attributes[TAX_ID] = '9606'
        self.data.attributes[GENE_AS_ATTRIBUTE_NAME] = True
        self.data.attributes[GENE_ID_ATTRIBUTE] = 'Entrez ID'

        self.genes = Table.from_list(
            Domain([], metas=[StringVariable('Entrez ID')]),
            [[attr.attributes.get('Entrez ID')] for attr in self.data.domain.attributes],
        )

        self.genes.attributes[TAX_ID] = '9606'
        self.genes.attributes[GENE_AS_ATTRIBUTE_NAME] = False
        self.genes.attributes[GENE_ID_COLUMN] = 'Entrez ID'

        self.expected_score_values = np.array([[3.0], [4.0], [1.6], [0]])

    def test_score_cells(self):
        # input data
        self.send_signal(self.widget.Inputs.data, self.data)
        self.assertIsNotNone(self.widget.input_data)
        self.assertIsNotNone(self.widget.input_genes)

        # marker data
        self.send_signal(self.widget.Inputs.genes, self.genes)

        self.assertIsNotNone(self.widget.marker_data)
        self.assertIsNotNone(self.widget.marker_genes)

        # output data
        self.widget.commit()
        out_data = self.get_output(self.widget.Outputs.data)
        self.assertTrue((self.expected_score_values == out_data.metas).all())

    def test_no_input_genes(self):
        # input data
        self.send_signal(self.widget.Inputs.data, self.data)
        self.assertIsNotNone(self.widget.input_data)

        # output data
        self.widget.commit()
        out_data = self.get_output(self.widget.Outputs.data)
        self.assertTrue(np.all(out_data.get_column_view('Score')[0] == 1.0))


if __name__ == '__main__':
    unittest.main()



from unittest.mock import patch, Mock

import numpy as np

from Orange.data import Table, Domain, DiscreteVariable
from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.tests.utils import simulate

from orangecontrib.single_cell.widgets.owdotmatrix import OWDotMatrix


class TestOWDotMatrix(WidgetTest):

    def setUp(self) -> None:
        self.iris = Table("iris")
        self.widget = self.create_widget(OWDotMatrix)

    def test_set_data(self):
        w = self.widget

        # send iris - dataset where analysis should pass
        self.send_signal(w.Inputs.data, self.iris)

        self.assertIsNotNone(w.data)
        self.assertIsNotNone(w.cluster_var)
        self.assertFalse(w.Error.no_discrete_variable.is_shown())

        # send housing - dataset where analysis should fail - no discrete var
        self.send_signal(w.Inputs.data, Table("housing"))

        self.assertIsNone(w.data)
        self.assertIsNone(w.cluster_var)
        self.assertTrue(w.Error.no_discrete_variable.is_shown())

        # reset input
        self.send_signal(w.Inputs.data, None)

        self.assertIsNone(w.data)
        self.assertIsNone(w.cluster_var)
        self.assertFalse(w.Error.no_discrete_variable.is_shown())

    def test_data_aggregation(self):
        w = self.widget

        # send iris - dataset where analysis should pass
        self.send_signal(w.Inputs.data, self.iris)
        self.assertEqual(3, len(w.clusters_unordered))
        self.assertTupleEqual((3, 4), w.aggregated_data.shape)

        self.send_signal(w.Inputs.data, None)
        self.assertIsNone(w.clusters_unordered)
        self.assertIsNone(w.aggregated_data)

    def test_data_aggregation_change_attr(self):
        """
        With this test we check whether discrete attribute for aggregation changes correctly
        """
        w = self.widget

        # add additional discrete attribute to data
        iris = Table(
            Domain(
                self.iris.domain.attributes,
                self.iris.domain.class_var,
                metas=[DiscreteVariable("a", values=["a", "b"])]),
            self.iris.X,
            self.iris.Y,
            metas=[[0]] * 100 + [[1]] * 50
        )

        # send iris - dataset where analysis should pass
        self.send_signal(w.Inputs.data, iris)
        cbox = self.widget.controls.cluster_var
        simulate.combobox_activate_index(cbox, 0)

        self.assertEqual(3, len(w.clusters_unordered))
        self.assertTupleEqual((3, 4), w.aggregated_data.shape)

        cbox = self.widget.controls.cluster_var
        simulate.combobox_activate_index(cbox, 2)
        self.assertEqual(2, len(w.clusters_unordered))
        self.assertTupleEqual((2, 4), w.aggregated_data.shape)

    def test_info_string(self):
        w = self.widget
        insum = w.info.set_input_summary = Mock()
        self.send_signal(w.Inputs.data, self.iris)
        self.assertEqual(insum.call_args[0][0], "150")
        self.assertEqual(insum.call_args[0][1], "4 genes\n150 cells\n3 clusters")

        self.send_signal(w.Inputs.data, None)
        self.assertEqual(insum.call_args[0][0], w.info.NoInput)

    def test_calculate_table_values_no_norm(self):
        w = self.widget
        w.biclustering = False
        w.transpose = False
        w.log_scale = False
        w.normalize = False

        self.send_signal(w.Inputs.data, self.iris)
        self.assertEqual(Table, type(w.matrix))
        np.testing.assert_almost_equal(w.matrix.X, w._norm_min_max(w.matrix_before_norm.X))
        np.testing.assert_almost_equal(w.matrix.X, w._norm_min_max(w.aggregated_data))
        np.testing.assert_almost_equal(w.matrix_before_norm.X, w.aggregated_data)
        self.assertListEqual(sorted(w.clusters), sorted(self.iris.domain.class_var.values))

    def test_calculate_table_values_norm(self):
        w = self.widget
        w.biclustering = False
        w.transpose = False
        w.log_scale = False
        w.normalize = True

        self.send_signal(w.Inputs.data, self.iris)
        self.assertEqual(Table, type(w.matrix))
        self.assertTupleEqual(w.matrix.X.shape, w.matrix_before_norm.X.shape)
        self.assertTupleEqual(w.matrix.X.shape, w.aggregated_data.shape)
        self.assertListEqual(sorted(w.clusters), sorted(self.iris.domain.class_var.values))

    def test_calculate_table_values_log(self):
        w = self.widget
        w.biclustering = False
        w.transpose = False
        w.log_scale = True
        w.normalize = False

        self.send_signal(w.Inputs.data, self.iris)
        self.assertEqual(Table, type(w.matrix))
        self.assertTupleEqual(w.matrix.X.shape, w.matrix_before_norm.X.shape)
        self.assertTupleEqual(w.matrix.X.shape, w.aggregated_data.shape)
        self.assertListEqual(sorted(w.clusters), sorted(self.iris.domain.class_var.values))

    def test_calculate_table_values_clustering(self):
        w = self.widget
        w.biclustering = True
        w.transpose = False
        w.log_scale = False
        w.normalize = False

        self.send_signal(w.Inputs.data, self.iris)
        self.assertEqual(Table, type(w.matrix))
        self.assertTupleEqual(w.matrix.X.shape, w.matrix_before_norm.X.shape)
        self.assertTupleEqual(w.matrix.X.shape, w.aggregated_data.shape)
        self.assertListEqual(sorted(w.clusters), sorted(self.iris.domain.class_var.values))
        self.assertListEqual(sorted(map(str, w.matrix.domain.attributes)),
                             sorted(map(str, self.iris.domain.attributes)))

    def test_calculate_table_values_transpose(self):
        w = self.widget
        w.biclustering = False
        w.transpose = True
        w.log_scale = False
        w.normalize = False

        self.send_signal(w.Inputs.data, self.iris)
        self.assertEqual(Table, type(w.matrix))
        self.assertTupleEqual(w.matrix.X.shape, w.matrix_before_norm.X.shape)
        self.assertTupleEqual(w.matrix.X.shape, w.aggregated_data.T.shape)
        self.assertListEqual(sorted(w.clusters), sorted(map(str, self.iris.domain.attributes)))
        self.assertListEqual(sorted(map(str, w.matrix.domain.attributes)),
                             sorted(self.iris.domain.class_var.values))

    def test_output(self):
        w = self.widget

        self.send_signal(w.Inputs.data, self.iris)
        ann_data = self.get_output(w.Outputs.annotated_data)
        sel_data = self.get_output(w.Outputs.selected_data)
        cont_data = self.get_output(w.Outputs.contingency)

        self.assertIsNotNone(ann_data)
        self.assertIsNone(sel_data)
        self.assertIsNotNone(cont_data)

        np.testing.assert_almost_equal(ann_data.X, w.data.X)
        np.testing.assert_almost_equal(ann_data.Y, w.data.Y)
        self.assertEqual(1, len(ann_data.domain.metas))
        np.testing.assert_almost_equal(cont_data.X, w.matrix.X)
        self.assertEqual(1, len(cont_data.domain.metas))
        self.assertEqual(1, len(cont_data.domain.metas))
        self.assertEqual("iris", str(cont_data.domain.metas[0]))

        self.send_signal(w.Inputs.data, None)

        ann_data = self.get_output(w.Outputs.annotated_data)
        sel_data = self.get_output(w.Outputs.selected_data)
        cont_data = self.get_output(w.Outputs.contingency)

        self.assertIsNone(ann_data)
        self.assertIsNone(sel_data)
        self.assertIsNone(cont_data)

    def test_output_transpose(self):
        w = self.widget

        self.send_signal(w.Inputs.data, self.iris)
        w.controls.transpose.click()

        ann_data = self.get_output(w.Outputs.annotated_data)
        sel_data = self.get_output(w.Outputs.selected_data)
        cont_data = self.get_output(w.Outputs.contingency)

        self.assertIsNotNone(ann_data)
        self.assertIsNone(sel_data)
        self.assertIsNotNone(cont_data)

        np.testing.assert_almost_equal(ann_data.X, w.data.X)
        np.testing.assert_almost_equal(ann_data.Y, w.data.Y)
        self.assertEqual(1, len(ann_data.domain.metas))
        np.testing.assert_almost_equal(cont_data.X, w.matrix.X)
        self.assertEqual(1, len(cont_data.domain.metas))
        self.assertEqual(1, len(cont_data.domain.metas))
        self.assertEqual("Gene", str(cont_data.domain.metas[0]))



import sys
import enum
import math
import numbers

from contextlib import contextmanager
from types import SimpleNamespace

import typing
from typing import Optional, Sequence, Tuple, Dict, Callable, Union, Iterable

import numpy as np
from scipy import stats

from AnyQt.QtCore import Qt, QSize, QPointF, QRectF, QLineF, QTimer
from AnyQt.QtCore import pyqtSignal as Signal, pyqtSlot as Slot
from AnyQt.QtGui import (
    QPainter, QPolygonF, QPainterPath, QPalette, QPen, QBrush, QColor,
    QKeySequence
)
from AnyQt.QtWidgets import (
    QLabel, QDoubleSpinBox, QGroupBox, QHBoxLayout, QAction,
    QGraphicsPathItem, QGraphicsRectItem, QGraphicsItem, QFormLayout,
    QApplication, QButtonGroup, QRadioButton, QCheckBox, QStackedWidget
)

import pyqtgraph as pg

import Orange.data
import Orange.widgets.utils.plot.owpalette
from Orange.widgets import widget, gui, settings

#: Filter type
Cells, Genes, Data = 0, 1, 2

#: Filter quality control measure (apply to Cell/Genes type only)
DetectionCount = 0  # number of genes/features with non-zero expression level
TotalCounts = 1  # total counts by cell/gene


#: Filter descriptions for various roles in UI
#: (short name, name, description)
FilterInfo = {
    Cells: ("Cells", "Cell Filter",
            "Filter cells (rows) by total counts (library size) or number "
            "of expressed genes."),
    Genes: ("Genes", "Gene Filter",
            "Filter genes (columns) by total counts mapped to a gene or "
            "number of cells in which the gene is expressed in."),
    Data: ("Data", "Data filter",
           "Filter out (zero) small measurements")
}

# Quality control measure descriptions for UI
MeasureInfo = {
    TotalCounts: ("Total counts",
                  "Sum of all counts across cell/gene"),
    DetectionCount: ("Detection count",
                     "Number of cells/genes with non-zero expression")
}


# Plot scales
class Scale(enum.Enum):
    Linear = "Linear"
    Log1p = "Log1p"


def log1p(x, base=10.):
    x = np.log1p(x)
    x /= np.log(base)
    return x


def expm1(x, base=10.):
    x = np.asarray(x)
    x *= np.log(base)
    return np.expm1(x)


class ScatterPlotItem(pg.ScatterPlotItem):
    def paint(self, painter, *args):
        if self.opts["antialias"]:
            painter.setRenderHint(QPainter.Antialiasing, True)
        if self.opts["pxMode"]:
            painter.setRenderHint(QPainter.SmoothPixmapTransform, True)
        super().paint(painter, *args)


if typing.TYPE_CHECKING:
    ArrayLike = Union[np.ndarray, np.generic, numbers.Number, Iterable]
    Transform = Callable[[ArrayLike], np.ndarray]


# filter state data class
class _FilterData:
    #: The array used for filtering/ploting. Is some marginal statistic
    #: of the input.
    x = ...   # type: np.ndarray
    #: The transformed array x (log1p).
    #: NOTE: xt.size need not match x.size (non finite values can be omitted)
    xt = ...  # type: np.ndarray
    #: The transformation function mapping x to xt
    transform = ...      # type: Transform
    #: The inverse of `transform`,
    transform_inv = ...  # type: Transform
    #: min/max bounds of x
    xmin = xmax = ...    # type: float
    #: min/max bounds of xt
    xtmin = xtmax = ...  # type: float


class OWFilter(widget.OWWidget):
    name = "Filter"
    icon = 'icons/Filter.svg'
    description = "Filter cells/genes"
    priority = 210

    class Inputs:
        data = widget.Input("Data", Orange.data.Table)

    class Outputs:
        data = widget.Output("Data", Orange.data.Table)

    class Warning(widget.OWWidget.Warning):
        sampling_in_effect = widget.Msg(
            "Too many data points to display.\n"
            "Sampling {} of {} data points."
        )

    class Error(widget.OWWidget.Error):
        invalid_range = widget.Msg(
            "Negative values in input data.\n"
            "This filter is only defined for non-negative values."
        )
        invalid_domain = widget.Msg(
            "Invalid domain\n"
            "Domain contains non numeric columns."
        )

    #: Filter mode.
    #: Filter out rows/columns or 'zap' data values in range.
    Cells, Genes, Data = Cells, Genes, Data

    settings_version = 3

    #: The selected filter mode
    selected_filter_type = settings.Setting(Cells)  # type: int

    #: Selected filter statistics / QC measure indexed by filter_type
    selected_filter_metric = settings.Setting(TotalCounts)  # type: int

    #: Augment the violin plot with a dot plot (strip plot) of the (non-zero)
    #: measurement counts in Cells/Genes mode or data matrix values in Data
    #: mode.
    display_dotplot = settings.Setting(True)  # type: bool

    #: Is min/max range selection enable
    limit_lower_enabled = settings.Setting(True)  # type: bool
    limit_upper_enabled = settings.Setting(True)  # type: bool

    #: The lower and upper selection limit for each filter type
    thresholds = settings.Setting({
        (Cells, DetectionCount): (0, 2 ** 31 - 1),
        (Cells, TotalCounts): (0, 2 ** 31 - 1),
        (Genes, DetectionCount): (0, 2 ** 31 - 1),
        (Genes, TotalCounts): (0, 2 ** 31 - 1),
        (Data, -1): (0.0, 2.0 ** 31 - 1)
    })  # type: Dict[Tuple[int, int], Tuple[float, float]]

    #: Plot scale: 'Linear' or 'Log1p'
    scale = settings.Setting(Scale.Linear.name)  # type: str

    auto_commit = settings.Setting(True)   # type: bool

    def __init__(self):
        super().__init__()
        self.data = None      # type: Optional[Orange.data.Table]
        self._state = None     # type: Optional[_FilterData]

        box = gui.widgetBox(self.controlArea, "Info")
        self._info = QLabel(box, wordWrap=True)
        self._info.setText("No data in input\n")

        box.layout().addWidget(self._info)

        box = gui.widgetBox(self.controlArea, "Filter Type", spacing=-1)
        rbg = QButtonGroup(box, exclusive=True)
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        for id_ in [Cells, Genes, Data]:
            name, _, tip = FilterInfo[id_]
            b = QRadioButton(
                name, toolTip=tip, checked=id_ == self.selected_filter_type
            )
            rbg.addButton(b, id_)
            layout.addWidget(b, stretch=10, alignment=Qt.AlignCenter)
        box.layout().addLayout(layout)

        rbg.buttonClicked[int].connect(self.set_filter_type)

        self.filter_metric_cb = gui.comboBox(
            box, self, "selected_filter_metric", callback=self._update_metric,
            enabled=self.selected_filter_type != Data
        )
        for id_ in [DetectionCount, TotalCounts]:
            text, ttip = MeasureInfo[id_]
            self.filter_metric_cb.addItem(text)
            idx = self.filter_metric_cb.count() - 1
            self.filter_metric_cb.setItemData(idx, ttip, Qt.ToolTipRole)
        self.filter_metric_cb.setCurrentIndex(self.selected_filter_metric)

        form = QFormLayout(
            labelAlignment=Qt.AlignLeft,
            formAlignment=Qt.AlignLeft,
            fieldGrowthPolicy=QFormLayout.AllNonFixedFieldsGrow
        )
        self._filter_box = box = gui.widgetBox(
            self.controlArea, "Filter", orientation=form
        )  # type: QGroupBox

        self.threshold_stacks = (
            QStackedWidget(enabled=self.limit_lower_enabled),
            QStackedWidget(enabled=self.limit_upper_enabled),
        )
        finfo = np.finfo(np.float64)
        for filter_ in [Cells, Genes, Data]:
            if filter_ in {Cells, Genes}:
                minimum = 0.0
                ndecimals = 1
                metric = self.selected_filter_metric
            else:
                minimum = finfo.min
                ndecimals = 3
                metric = -1
            spinlower = QDoubleSpinBox(
                self, minimum=minimum, maximum=finfo.max, decimals=ndecimals,
                keyboardTracking=False,
            )
            spinupper = QDoubleSpinBox(
                self, minimum=minimum, maximum=finfo.max, decimals=ndecimals,
                keyboardTracking=False,
            )

            lower, upper = self.thresholds.get((filter_, metric), (0, 0))

            spinlower.setValue(lower)
            spinupper.setValue(upper)

            self.threshold_stacks[0].addWidget(spinlower)
            self.threshold_stacks[1].addWidget(spinupper)

            spinlower.valueChanged.connect(self._limitchanged)
            spinupper.valueChanged.connect(self._limitchanged)

        self.threshold_stacks[0].setCurrentIndex(self.selected_filter_type)
        self.threshold_stacks[1].setCurrentIndex(self.selected_filter_type)

        self.limit_lower_enabled_cb = cb = QCheckBox(
            "Min", checked=self.limit_lower_enabled
        )
        cb.toggled.connect(self.set_lower_limit_enabled)
        cb.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)
        form.addRow(cb, self.threshold_stacks[0])

        self.limit_upper_enabled_cb = cb = QCheckBox(
            "Max", checked=self.limit_upper_enabled
        )
        cb.toggled.connect(self.set_upper_limit_enabled)
        cb.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)
        form.addRow(cb, self.threshold_stacks[1])

        box = gui.widgetBox(self.controlArea, "Plot Options")
        self._showpoints = gui.checkBox(
            box, self, "display_dotplot", "Show data points",
            callback=self._update_dotplot
        )
        self.log_scale_cb = QCheckBox(
            "Log scale", checked=self.scale == Scale.Log1p.name
        )
        self.log_scale_cb.toggled[bool].connect(
            lambda state:
                self.set_filter_scale(Scale.Log1p if state else Scale.Linear)
        )
        box.layout().addWidget(self.log_scale_cb)

        self.controlArea.layout().addStretch(10)

        gui.auto_commit(self.controlArea, self, "auto_commit", "Commit")

        self._view = pg.GraphicsView()
        self._view.enableMouse(False)
        self._view.setAntialiasing(True)
        self._plot = plot = ViolinPlot()
        self._plot.setDataPointsVisible(self.display_dotplot)
        self._plot.setSelectionMode(
            (ViolinPlot.Low if self.limit_lower_enabled else 0) |
            (ViolinPlot.High if self.limit_upper_enabled else 0)
        )
        self._plot.setRange(QRectF(-1., 0., 2., 1.))
        self._plot.selectionEdited.connect(self._limitchanged_plot)
        self._view.setCentralWidget(self._plot)

        bottom = self._plot.getAxis("bottom")  # type: pg.AxisItem
        bottom.hide()
        plot.setMouseEnabled(False, False)
        plot.hideButtons()
        self.mainArea.layout().addWidget(self._view)

        # Coalescing commit timer
        self._committimer = QTimer(self, singleShot=True)
        self._committimer.timeout.connect(self.commit)

        self.addAction(
            QAction("Select All", self, shortcut=QKeySequence.SelectAll,
                    triggered=self._select_all)
        )
        self._setup_axes()

    def sizeHint(self):
        sh = super().sizeHint()  # type: QSize
        return sh.expandedTo(QSize(800, 600))

    def set_filter_type(self, type_):
        if self.selected_filter_type != type_:
            assert type_ in (Cells, Genes, Data), str(type_)
            self.selected_filter_type = type_
            self.threshold_stacks[0].setCurrentIndex(type_)
            self.threshold_stacks[1].setCurrentIndex(type_)
            self.filter_metric_cb.setEnabled(type_ != Data)
            self._setup_axes()
            if self.data is not None:
                self._setup(self.data, type_)
                self._schedule_commit()

    def filter_type(self):
        return self.selected_filter_type

    def set_filter_scale(self, scale):
        # type: (Scale) -> None
        if self.scale != scale:
            self.scale = scale.name
            self.log_scale_cb.setChecked(scale == Scale.Log1p)
            self._update_scale()

    def filter_scale(self):
        return Scale[self.scale]

    def _update_metric(self):
        self._update_scale()
        if self.data is not None:
            self._setup(self.data, self.selected_filter_type, )

    def set_upper_limit_enabled(self, enabled):
        if enabled != self.limit_upper_enabled:
            self.limit_upper_enabled = enabled
            self.threshold_stacks[1].setEnabled(enabled)
            self.limit_upper_enabled_cb.setChecked(enabled)
            self._update_filter()
            self._schedule_commit()

    def set_lower_limit_enabled(self, enabled):
        if enabled != self.limit_lower_enabled:
            self.limit_lower_enabled = enabled
            self.threshold_stacks[0].setEnabled(enabled)
            self.limit_lower_enabled_cb.setChecked(enabled)
            self._update_filter()
            self._schedule_commit()

    def _update_filter(self):
        mode = 0
        if self.limit_lower_enabled:
            mode |= ViolinPlot.Low
        if self.limit_upper_enabled:
            mode |= ViolinPlot.High
        self._plot.setSelectionMode(mode)
        self._update_info()
        self._schedule_commit()

    def _is_filter_enabled(self):
        return self.limit_lower_enabled or self.limit_upper_enabled

    @Inputs.data
    def set_data(self, data):
        # type: (Optional[Orange.data.Table]) -> None
        self.clear()

        if data is not None and \
                any(type(v) is not Orange.data.ContinuousVariable
                    for v in data.domain.attributes):
            self.Error.invalid_domain()
            data = None

        if data is not None and np.any(data.X < 0):
            self.Error.invalid_range()
            data = None

        self.data = data

        if data is not None:
            self._setup(data, self.filter_type())

        self.unconditional_commit()

    def clear(self):
        self.data = None
        self._state = None
        self._plot.clear()
        # reset the plot range
        self._plot.setRange(QRectF(-1., 0., 2., 1.))
        self._update_info()
        self.Warning.clear()
        self.Error.clear()

    def _update_info(self):
        text = []
        if self.data is None:
            text += ["No data on input.\n"]
        else:
            N, M = len(self.data), len(self.data.domain.attributes)
            text = []
            text += [
                "Data with {N} cell{Np} and {M} gene{Mp}"
                .format(N=N, Np="s" if N != 1 else "",
                        M=M, Mp="s" if N != 1 else "")
            ]
            if self._is_filter_enabled() and \
                    self.filter_type() in [Cells, Genes]:
                counts = self._state.x
                mask = np.ones(counts.shape, dtype=bool)
                if self.limit_lower_enabled:
                    mask &= self.limit_lower <= counts

                if self.limit_upper_enabled:
                    mask &= counts <= self.limit_upper

                n = np.count_nonzero(mask)
                subject = "cell" if self.filter_type() == Cells else "gene"
                if n == 0:
                    text += ["All {}s filtered out".format(subject)]
                else:
                    text += [
                        "{} {subject}{s} in selection"
                        .format(n, subject=subject, s="s" if n != 1 else "")
                    ]
            else:
                text += [""]
        self._info.setText("\n".join(text))

    def _select_all(self):
        self.limit_lower = 0
        self.limit_upper = 2 ** 31 - 1
        self._limitchanged()

    def _setup_axes(self):
        # Setup the plot axes and title
        filter_type = self.filter_type()
        info = FilterInfo[filter_type]
        _, title, _, *_ = info

        if filter_type in [Cells, Genes]:
            measure = self.selected_filter_metric
        else:
            measure = None

        if filter_type == Cells and measure == TotalCounts:
            axis_label = "Total counts (library size)"
        elif filter_type == Cells and measure == DetectionCount:
            axis_label = "Number of expressed genes"
        elif filter_type == Genes and measure == TotalCounts:
            axis_label = "Total counts"
        elif filter_type == Genes and measure == DetectionCount:
            # TODO: Too long
            axis_label = "Number of cells a gene is expressed in"
        elif filter_type == Data:
            axis_label = "Gene Expression"

        ax = self._plot.getAxis("left")
        if self.filter_scale() == Scale.Log1p:
            axis_label = "1 + '{}' <i>(in log scale)</i>".format(axis_label)
            ax.setLabel(axis_label)
            ax.setLogMode(True)
        else:
            ax.setLogMode(False)
            ax.setLabel(axis_label)
        # Reset the tick text area width
        ax.textWidth = 30
        ax.setWidth(None)

        self._plot.setTitle(title)

    def _setup(self, data, filter_type):
        self._plot.clear()
        self._state = None
        self._setup_axes()

        span = -1.0  # data span
        measure = self.selected_filter_metric if filter_type != Data else None
        state = _FilterData()

        if filter_type in [Cells, Genes]:
            if filter_type == Cells:
                axis = 1
            else:
                axis = 0
            if measure == TotalCounts:
                counts = np.nansum(data.X, axis=axis)
            else:
                mask = (data.X != 0) & (np.isfinite(data.X))
                counts = np.count_nonzero(mask, axis=axis)
            x = counts
            self.Warning.sampling_in_effect.clear()
        elif filter_type == Data:
            x = data.X.ravel()
            x = x[np.isfinite(x)]
            x = x[x != 0]
            MAX_DISPLAY_SIZE = 20000
            if x.size > MAX_DISPLAY_SIZE:
                self.Warning.sampling_in_effect(MAX_DISPLAY_SIZE, x.size)
                # tails to preserve exactly
                tails = 1
                assert x.flags.owndata
                x.partition(tails - 1)
                xrest = x[tails:]
                xrest.partition(xrest.size - tails)

                x1, x2, x3 = x[:tails], x[tails:x.size - tails], x[x.size-tails:]
                assert x1.size + x2.size + x3.size == x.size
                x2 = np.random.RandomState(0x667).choice(
                    x2, size=MAX_DISPLAY_SIZE - 2 * tails, replace=False,
                )
                x = np.r_[x1, x2, x3]
            else:
                self.Warning.sampling_in_effect.clear()
        else:
            assert False

        state.x = x

        scale = self.filter_scale()

        if scale == Scale.Log1p:
            scale_transform = log1p
            scale_transform_inv = expm1
        else:
            scale_transform = lambda x: x
            scale_transform_inv = scale_transform

        state.transform = scale_transform
        state.transform_inv = scale_transform_inv

        if x.size:
            xmin, xmax = np.min(x), np.max(x)
        else:
            xmin = xmax = 0., 1.

        state.xmin, state.xmax = xmin, xmax

        xs = scale_transform(x)
        xs = xs[np.isfinite(xs)]
        state.xt = xs

        if xs.size:
            xsmin, xsmax = np.min(xs), np.max(xs)
            # find effective xmin, xmax (valid in both original and transformed
            # space
            xmin_, xmax_ = scale_transform_inv([xsmin, xsmax])
            xmin, xmax = max(xmin, xmin_), min(xmax, xmax_)

            lower = np.clip(self.limit_lower, xmin, xmax)
            upper = np.clip(self.limit_upper, xmin, xmax)
        else:
            xmin, xmax = 0., 1.
            lower, upper = 0., 1.

        state.xtmin, state.xtmax = xsmin, xsmax

        spinlow = self.threshold_stacks[0].widget(filter_type)
        spinhigh = self.threshold_stacks[1].widget(filter_type)
        if filter_type == Data or measure == TotalCounts:
            span = xmax - xmin
            if span > 0:
                ndecimals = max(4 - int(np.floor(np.log10(span))), 1)
            else:
                ndecimals = 1
        else:
            ndecimals = 1

        # Round effective bounds (spin <=> plot cut lines)
        lower = round(lower, ndecimals)
        upper = round(upper, ndecimals)

        if xs.size > 0:
            # TODO: Need correction for lower bounded distribution (counts)
            # Use reflection around 0, but gaussian_kde does not provide
            # sufficient flexibility w.r.t bandwidth selection.

            self._plot.setData(xs, 1000)
            self._plot.setBoundary(*scale_transform([lower, upper]))

        spinlow.setDecimals(ndecimals)
        self.limit_lower = lower

        spinhigh.setDecimals(ndecimals)
        self.limit_upper = upper

        self._state = state
        self._update_info()

    def _update_dotplot(self):
        self._plot.setDataPointsVisible(self.display_dotplot)

    def current_filter_thresholds(self):
        if self.selected_filter_type in {Cells, Genes}:
            metric = self.selected_filter_metric
        else:
            metric = -1
        return self.thresholds[self.selected_filter_type, metric]

    def set_current_filter_thesholds(self, lower, upper):
        if self.selected_filter_type in {Cells, Genes}:
            metric = self.selected_filter_metric
        else:
            metric = -1
        self.thresholds[self.selected_filter_type, metric] = (lower, upper)

    def _update_scale(self):
        self._setup_axes()
        if self.data is not None:
            self._setup(self.data, self.filter_type())

    @property
    def limit_lower(self):
        return self.current_filter_thresholds()[0]

    @limit_lower.setter
    def limit_lower(self, value):
        _, upper = self.current_filter_thresholds()
        self.set_current_filter_thesholds(value, upper)
        stacklower, _ = self.threshold_stacks
        sb = stacklower.widget(self.selected_filter_type)
        # prevent changes due to spin box rounding
        sb.setValue(value)

    @property
    def limit_upper(self):
        return self.current_filter_thresholds()[1]

    @limit_upper.setter
    def limit_upper(self, value):
        lower, _ = self.current_filter_thresholds()
        self.set_current_filter_thesholds(lower, value)
        _, stackupper = self.threshold_stacks
        sb = stackupper.widget(self.selected_filter_type)
        sb.setValue(value)

    @Slot()
    def _limitchanged(self):
        # Low/high limit changed via the spin boxes
        stacklow, stackhigh = self.threshold_stacks
        filter_ = self.selected_filter_type

        lower = stacklow.widget(filter_).value()
        upper = stackhigh.widget(filter_).value()
        self.set_current_filter_thesholds(lower, upper)

        state = self._state
        if state is not None and state.x.size:
            xmin, xmax = state.xmin, state.xmax
            lower = np.clip(lower, xmin, xmax)
            upper = np.clip(upper, xmin, xmax)
            lower, upper = state.transform([lower, upper])
            self._plot.setBoundary(lower, upper)
            # TODO: Only when the actual selection/filter mask changes
            self._schedule_commit()
            self._update_info()

    def _limitchanged_plot(self):
        # Low/high limit changed via the plot
        if self._state is not None:
            state = self._state
            newlower_, newupper_ = self._plot.boundary()
            newlower, newupper = state.transform_inv([newlower_, newupper_])
            filter_ = self.selected_filter_type
            lower, upper = self.current_filter_thresholds()
            stacklow, stackhigh = self.threshold_stacks
            spin_lower = stacklow.widget(filter_)
            spin_upper = stackhigh.widget(filter_)
            # do rounding to match the spin box's precision
            if self.limit_lower_enabled:
                newlower = round(newlower, spin_lower.decimals())
            else:
                newlower = lower

            if self.limit_upper_enabled:
                newupper = round(newupper, spin_upper.decimals())
            else:
                newupper = upper

            if self.limit_lower_enabled and newlower != lower:
                self.limit_lower = newlower
            if self.limit_upper_enabled and newupper != upper:
                self.limit_upper = newupper

            newlower_, newupper_ = state.transform([newlower, newupper])
            self._plot.setBoundary(newlower_, newupper_)

            # TODO: Only when the actual selection/filter mask changes
            self._schedule_commit()
            self._update_info()

    def _schedule_commit(self):
        self._committimer.start()

    def commit(self):
        self._committimer.stop()
        data = self.data

        if data is not None and self._is_filter_enabled():
            if self.filter_type() in [Cells, Genes]:
                state = self._state
                assert state is not None
                counts = state.x
                cmax = self.limit_upper
                cmin = self.limit_lower
                mask = np.ones(counts.shape, dtype=bool)
                if self.limit_lower_enabled:
                    mask &= cmin <= counts
                if self.limit_upper_enabled:
                    mask &= counts <= cmax

                if self.filter_type() == Cells:
                    assert counts.size == len(data)
                    data = data[mask]
                else:
                    assert counts.size == len(data.domain.attributes)
                    atts = [v for v, m in zip(data.domain.attributes, mask)
                            if m]
                    data = data.from_table(
                        Orange.data.Domain(
                            atts, data.domain.class_vars, data.domain.metas
                        ),
                        data
                    )
                if (
                    len(data) == 0
                    or len(data.domain.variables) + len(data.domain.metas) == 0
                ):
                    data = None
            elif self.filter_type() == Data:
                dmin, dmax = self.limit_lower, self.limit_upper
                data = data.copy()
                assert data.X.base is None
                mask = None
                if self.limit_lower_enabled:
                    mask = data.X < dmin
                if self.limit_upper_enabled:
                    if mask is not None:
                        mask |= data.X > dmax
                    else:
                        mask = data.X < dmax
                data.X[mask] = 0.0
            else:
                assert False

        self.Outputs.data.send(data)

    def onDeleteWidget(self):
        self.clear()
        self._plot.close()
        self._view.scene().clear()
        super().onDeleteWidget()

    @classmethod
    def migrate_settings(cls, settings, version):
        if (version is None or version < 2) and \
                ("limit_lower" in settings and "limit_upper" in settings):
            # v2 changed limit_lower, limit_upper to per filter limits stored
            # in a single dict
            lower = settings.pop("limit_lower")
            upper = settings.pop("limit_upper")
            settings["thresholds"] = {
                (Cells, TotalCounts): (lower, upper),
                (Cells, DetectionCount): (lower, upper),
                (Genes, TotalCounts): (lower, upper),
                (Genes, DetectionCount): (lower, upper),
                (Data, -1): (lower, upper),
            }
        if version == 2:
            thresholds = settings["thresholds"]
            c = thresholds.pop(Cells)
            g = thresholds.pop(Genes)
            d = thresholds.pop(Data)
            thresholds = {
                (Cells, TotalCounts): c,
                (Cells, DetectionCount): c,
                (Genes, TotalCounts): g,
                (Genes, DetectionCount): g,
                (Data, -1): d,
            }
            settings["thresholds"] = thresholds



@contextmanager
def block_signals(qobj):
    b = qobj.blockSignals(True)
    try:
        yield
    finally:
        qobj.blockSignals(b)


class AxisItem(pg.AxisItem):
    def logTickStrings(self, values, scale, spacing):
        # reimplemented
        values = [10 ** v for v in values]
        return [render_exp(v, 1) for v in values]


class ViolinPlot(pg.PlotItem):
    """
    A violin plot item with interactive data boundary selection.
    """
    #: Emitted when the selection boundary has changed
    selectionChanged = Signal()
    #: Emitted when the selection boundary has been edited by the user
    #: (by dragging the boundary lines)
    selectionEdited = Signal()

    #: Selection Flags
    NoSelection, Low, High = 0, 1, 2

    def __init__(self, *args, enableMenu=False, axisItems=None, **kwargs):
        if axisItems is None:
            axisItems = {}
        for position in ("left", 'right', 'top', 'bottom'):
            axisItems.setdefault(position, AxisItem(position))

        super().__init__(*args, enableMenu=enableMenu, axisItems=axisItems,
                         **kwargs)
        self.__data = None
        #: min/max cutoff line positions
        self.__min = 0
        self.__max = 0
        self.__dataPointsVisible = True
        self.__selectionEnabled = True
        self.__selectionMode = ViolinPlot.High | ViolinPlot.Low
        self._plotitems = None

    def setData(self, data, nsamples, sample_range=None, color=Qt.magenta):
        assert np.all(np.isfinite(data))

        if data.size > 0:
            xmin, xmax = np.min(data), np.max(data)
        else:
            xmin = xmax = 0.0

        if sample_range is None:
            xrange = xmax - xmin
            sample_min = xmin - xrange * 0.025
            sample_max = xmax + xrange * 0.025
        else:
            sample_min, sample_max = sample_range

        sample = np.linspace(sample_min, sample_max, nsamples)
        if data.size < 2:
            est = np.full(sample.size, 1. / sample.size, )
        else:
            try:
                density = stats.gaussian_kde(data)
                est = density.evaluate(sample)
            except np.linalg.LinAlgError:
                est = np.zeros(sample.size)

        item = QGraphicsPathItem(violin_shape(sample, est))
        color = QColor(color)
        color.setAlphaF(0.5)
        item.setBrush(QBrush(color))
        pen = QPen(self.palette().color(QPalette.Shadow))
        pen.setCosmetic(True)
        item.setPen(pen)
        est_max = np.max(est)

        x = np.random.RandomState(0xD06F00D).uniform(
            -est_max, est_max, size=data.size
        )
        dots = ScatterPlotItem(
            x=x, y=data, size=3,
        )
        dots.setVisible(self.__dataPointsVisible)
        pen = QPen(self.palette().color(QPalette.Shadow), 1)
        hoverPen = QPen(self.palette().color(QPalette.Highlight), 1.5)
        cmax = SelectionLine(
            angle=0, pos=xmax, movable=True, bounds=(sample_min, sample_max),
            pen=pen, hoverPen=hoverPen
        )
        cmin = SelectionLine(
            angle=0, pos=xmin, movable=True, bounds=(sample_min, sample_max),
            pen=pen, hoverPen=hoverPen
        )
        cmax.setCursor(Qt.SizeVerCursor)
        cmin.setCursor(Qt.SizeVerCursor)

        selection_item = QGraphicsRectItem(
            QRectF(-est_max, xmin, est_max * 2, xmax - xmin)
        )
        selection_item.setPen(QPen(Qt.NoPen))
        selection_item.setBrush(QColor(0, 250, 0, 50))

        def update_selection_rect():
            mode = self.__selectionMode
            p = selection_item.parentItem()  # type: Optional[QGraphicsItem]
            while p is not None and not isinstance(p, pg.ViewBox):
                p = p.parentItem()
            if p is not None:
                viewbox = p  # type: pg.ViewBox
            else:
                viewbox = None
            rect = selection_item.rect()  # type: QRectF
            if mode & ViolinPlot.High:
                rect.setTop(cmax.value())
            elif viewbox is not None:
                rect.setTop(viewbox.viewRect().bottom())
            else:
                rect.setTop(cmax.maxRange[1])

            if mode & ViolinPlot.Low:
                rect.setBottom(cmin.value())
            elif viewbox is not None:
                rect.setBottom(viewbox.viewRect().top())
            else:
                rect.setBottom(cmin.maxRange[0])

            selection_item.setRect(rect.normalized())

        cmax.sigPositionChanged.connect(update_selection_rect)
        cmin.sigPositionChanged.connect(update_selection_rect)
        cmax.visibleChanged.connect(update_selection_rect)
        cmin.visibleChanged.connect(update_selection_rect)

        def setupper(line):
            ebound = self.__effectiveBoundary()
            elower, eupper = ebound
            mode = self.__selectionMode
            if not mode & ViolinPlot.High:
                return
            upper = line.value()
            lower = min(elower, upper)
            if lower != elower and mode & ViolinPlot.Low:
                self.__min = lower
                cmin.setValue(lower)

            if upper != eupper:
                self.__max = upper

            if ebound != self.__effectiveBoundary():
                self.selectionEdited.emit()
                self.selectionChanged.emit()

        def setlower(line):
            ebound = self.__effectiveBoundary()
            elower, eupper = ebound
            mode = self.__selectionMode
            if not mode & ViolinPlot.Low:
                return
            lower = line.value()
            upper = max(eupper, lower)
            if upper != eupper and mode & ViolinPlot.High:
                self.__max = upper
                cmax.setValue(upper)

            if lower != elower:
                self.__min = lower

            if ebound != self.__effectiveBoundary():
                self.selectionEdited.emit()
                self.selectionChanged.emit()

        cmax.sigPositionChanged.connect(setupper)
        cmin.sigPositionChanged.connect(setlower)
        selmode = self.__selectionMode
        cmax.setVisible(selmode & ViolinPlot.High)
        cmin.setVisible(selmode & ViolinPlot.Low)
        selection_item.setVisible(selmode)

        self.addItem(dots)
        self.addItem(item)
        self.addItem(cmax)
        self.addItem(cmin)
        self.addItem(selection_item)

        self.setRange(
            QRectF(-est_max, np.min(sample), est_max * 2, np.ptp(sample))
        )
        self._plotitems = SimpleNamespace(
            pointsitem=dots,
            densityitem=item,
            cmax=cmax,
            cmin=cmin,
            selection_item=selection_item
        )
        self.__min = xmin
        self.__max = xmax

    def setDataPointsVisible(self, visible):
        self.__dataPointsVisible = visible
        if self._plotitems is not None:
            self._plotitems.pointsitem.setVisible(visible)

    def setSelectionMode(self, mode):
        oldlower, oldupper = self.__effectiveBoundary()
        oldmode = self.__selectionMode
        mode = mode & 0b11
        if self.__selectionMode == mode:
            return

        self.__selectionMode = mode

        if self._plotitems is None:
            return

        cmin = self._plotitems.cmin
        cmax = self._plotitems.cmax
        selitem = self._plotitems.selection_item

        cmin.setVisible(mode & ViolinPlot.Low)
        cmax.setVisible(mode & ViolinPlot.High)
        selitem.setVisible(bool(mode))

        lower, upper = self.__effectiveBoundary()
        # The recorded values are not bounded by each other on gui interactions
        # when one is disabled. Rectify this now.
        if (oldmode ^ mode) & ViolinPlot.Low and mode & ViolinPlot.High:
            # Lower activated and High enabled
            lower = min(lower, upper)
        if (oldmode ^ mode) & ViolinPlot.High and mode & ViolinPlot.Low:
            # High activated and Low enabled
            upper = max(lower, upper)

        with block_signals(self):
            if lower != oldlower and mode & ViolinPlot.Low:
                cmin.setValue(lower)
            if upper != oldupper and mode & ViolinPlot.High:
                cmax.setValue(upper)

        self.selectionChanged.emit()

    def setBoundary(self, low, high):
        """
        Set the lower and upper selection boundary value.
        """
        changed = 0
        mode = self.__selectionMode
        if self.__min != low:
            self.__min = low
            changed |= mode & ViolinPlot.Low
        if self.__max != high:
            self.__max = high
            changed |= mode & ViolinPlot.High

        if changed:
            if self._plotitems:
                with block_signals(self):
                    if changed & ViolinPlot.Low:
                        self._plotitems.cmin.setValue(low)
                    if changed & ViolinPlot.High:
                        self._plotitems.cmax.setValue(high)

            self.selectionChanged.emit()

    def boundary(self):
        """
        Return the current lower and upper selection boundary values.
        """
        return self.__min, self.__max

    def __effectiveBoundary(self):
        # effective boundary, masked by selection mode
        low, high = -np.inf, np.inf
        if self.__selectionMode & ViolinPlot.Low:
            low = self.__min
        if self.__selectionMode & ViolinPlot.High:
            high = self.__max
        return low, high

    def clear(self):
        super().clear()
        self._plotitems = None

    def mouseDragEvent(self, event):
        mode = self.__selectionMode
        if mode != ViolinPlot.NoSelection and event.buttons() & Qt.LeftButton:
            start = event.buttonDownScenePos(Qt.LeftButton)  # type: QPointF
            pos = event.scenePos()  # type: QPointF
            cmin, cmax = self._plotitems.cmin, self._plotitems.cmax
            assert cmin.parentItem() is cmax.parentItem()
            pos = self.mapToItem(cmin.parentItem(), pos)
            start = self.mapToItem(cmin.parentItem(), start)
            if mode & ViolinPlot.Low and mode & ViolinPlot.High:
                lower, upper = min(pos.y(), start.y()), max(pos.y(), start.y())
                cmin.setValue(lower)
                cmax.setValue(upper)
            elif mode & ViolinPlot.Low:
                lower = pos.y()
                cmin.setValue(lower)
            elif mode & ViolinPlot.High:
                upper = pos.y()
                cmax.setValue(upper)
            event.accept()


def violin_shape(x, p):
    # type: (Sequence[float], Sequence[float]) -> QPainterPath
    points = [QPointF(pi, xi) for xi, pi in zip(x, p)]
    points += [QPointF(-pi, xi) for xi, pi in reversed(list(zip(x, p)))]
    poly = QPolygonF(points)
    path = QPainterPath()
    path.addPolygon(poly)
    return path


class SelectionLine(pg.InfiniteLine):
    def paint(self, painter, option, widget=None):
        brect = self.boundingRect()
        c = brect.center()
        line = QLineF(brect.left(), c.y(), brect.right(), c.y())
        t = painter.transform()
        line = t.map(line)
        painter.save()
        painter.resetTransform()
        painter.setPen(self.currentPen)
        painter.drawLine(line)
        painter.restore()


def render_exp(value, prec=2):
    # type: (float, int) -> str
    if not math.isfinite(value):
        return repr(value)
    exp = "{:.{prec}G}".format(value, prec=prec)
    try:
        frac, exp = exp.split("E", 1)
    except ValueError:
        return exp

    frac = float(frac)
    exp = int(exp)
    if exp == 0:
        return str(frac)
    elif frac == 1.0:
        return "10{exp}".format(exp=_superscript(str(exp)))
    else:
        return "{frac:g}\u00D710{exp}".format(
            frac=frac, exp=_superscript(str(exp))
        )


def _superscript(string):
    # type: (str) -> str
    table = str.maketrans(
        "0123456789+-",
        "\u2070\u00B9\u00B2\u00B3\u2074\u2075\u2076\u2077\u2078\u2079"
        "\u207A\u207B",
    )
    return string.translate(table)


def main(argv=None):  # pragma: no cover
    app = QApplication(list(argv or sys.argv))
    argv = app.arguments()
    w = OWFilter()
    if len(argv) > 1:
        filename = argv[1]
        data = Orange.data.Table(filename)
    else:
        X = np.random.exponential(size=(1000, 1050)) - 1
        X[X < 0] = 0
        data = Orange.data.Table.from_numpy(None, X)
    w.set_data(data)
    w.show()
    w.raise_()
    app.exec()
    w.saveSettings()
    w.onDeleteWidget()


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main(sys.argv))



import numpy as np
import vcf

from Orange.data import DiscreteVariable, Domain, StringVariable, Table


class VariantData:
    def __init__(self, filename):
        reader = vcf.Reader(open(filename, "r"))
        records = [r for r in reader]

        self.samples = np.array(reader.samples)

        self.gq = np.array([[s.data.GQ for s in r.samples] for r in records],
                           dtype="f")
        self.gq = np.nan_to_num(self.gq)

        gt = np.array([[s.data.GT for s in r.samples] for r in records])
        self.gt = gt != "0/0"
        self.records = records

        self.variables = [
            DiscreteVariable("%s-%s" % (r.CHROM, r.POS), values=["0", "1"])
            for r in self.records
        ]

        for v, r in zip(self.variables, records):
            v.attributes["CHROM"] = str(r.CHROM)
            v.attributes["POS"] = str(r.POS)
            v.attributes["REF"] = str(r.REF)
            v.attributes["ALT"] = "".join(str(s) for s in r.ALT)

    def info(self):
        print("Samples: %d" % len(self.samples))
        unique_samples = set(s[:-2] for s in self.samples)
        print("Unique samples: %d" % len(unique_samples))
        print("Variants: %d" % len(self.records))

    def get_data(self, quality=None, frequency=None):
        """Orange data table with genotypes above quality and frequency
        threshold."""

        X = self.gt.astype(dtype="float", copy=True)
        if quality is not None:
            X[self.gq < quality] = np.nan
        selected = ~np.isnan(X).all(axis=1)
        if frequency is not None:
            selected &= np.nansum(X, axis=1) >= frequency
        X = X[selected].T

        variables = tuple(np.array(self.variables)[selected])
        metas = [StringVariable(s) for s in ["sample", "id"]]
        M = np.empty((len(X), 2), dtype="object")
        M[:, 0] = self.samples
        M[:, 1] = np.arange(1, len(X) + 1)
        domain = Domain(variables, [], metas=metas)
        data = Table(domain, X, Y=None, metas=M)

        return data

if __name__ == "__main__":
    variants = VariantData("cells.vcf")
    data = variants.get_data(30, 30)



import numpy as np
import itertools as it
from fastdtw import fastdtw
from scipy.stats import spearmanr, scoreatpercentile
from orangecontrib.single_cell.preprocess.biweight import biweight_midcorrelation
from orangecontrib.single_cell.preprocess.cca import MultiCCA, SVDCCA


GENE_SCORING_METHODS = ("pearson", "spearman", "bicor")


def score_genes(Xs, Ws, n_metagenes=30, method=GENE_SCORING_METHODS[0]):
    """ Score genes To construct a metagene.
        Each CC component gets its own set of characteristic genes. """
    n_datasets = len(Xs)
    n_genes = Xs[0].shape[1]
    n_cc = Ws[0].shape[1]
    use_genes = dict()

    # Each CC is scored independently
    for j in range(n_cc):
        gene_scores = np.zeros((n_genes, n_datasets))
        if method == "pearson":
            for d in range(n_datasets):
                Nx = np.linalg.norm(Xs[d], axis=0)
                gene_scores[:, d] = Ws[d][:, j].T.dot(Xs[d]) / Nx
        elif method == "spearman":
            for d, g in it.product(range(n_datasets), range(n_genes)):
                gene_scores[g, d] = spearmanr(Xs[d][:, g], Ws[d][:, j])[0]
        elif method == "bicor":
            for d, g in it.product(range(n_datasets), range(n_genes)):
                gene_scores[g, d] = biweight_midcorrelation(Xs[d][:, g], Ws[d][:, j])
        else:
            raise ValueError("Parameter method must be in {pearson, spearman, bicor}.")

        # Apply sorting criteria; high is good, low is bad
        f0 = np.isnan(gene_scores).sum(axis=1)
        f1 = - np.product(np.sign(gene_scores[:, 0:1]) == np.sign(gene_scores[:, 1:]), axis=1)
        f2 = - np.min(np.absolute(gene_scores), axis=1)
        order = sorted(range(n_genes), key=lambda i: (f0[i], f1[i], f2[i]))
        use_genes[j] = order[:n_metagenes]

    return use_genes


def reference_range(x, min_p=2.5, max_p=97.5):
    """ Map values in x to its (min_p), (max_p) quantile range. """
    qmin = scoreatpercentile(x, min_p)
    qmax = scoreatpercentile(x, max_p)
    return (x - qmin) / (qmax - qmin)


def quantile_shift(x, y):
    """ Correct y for quantile shifts with reference to x. """
    shift = np.min([abs(scoreatpercentile(x, z) - scoreatpercentile(y, z))
                    for z in range(10, 100, 10)])
    return y + shift


def metagene_map(Xs, Ws, use_genes, align=True, qmin=2.5, qmax=97.5):
    """ Linear map to the meta genes.
        Assumption: Xs (and Ws) are sorted on descending number of rows. """
    n_datasets = len(Xs)
    n_cc = Ws[0].shape[1]
    Phis = [np.zeros((X.shape[0], n_cc)) for X in Xs]

    # Project to the selected set of metagenes
    for j, genes in use_genes.items():
        for X, W, Phi in zip(Xs, Ws, Phis):
            Phi[:, j] = W[:, j].T.dot(X[:, genes]).dot(X[:, genes].T).T

        # Align all datasets to the largest one
        if align:
            Phis[0][:, j] = reference_range(Phis[0][:, j], min_p=qmin, max_p=qmax)
            for d in range(1, n_datasets):
                Phis[d][:, j] = quantile_shift(Phis[0][:, j],
                                               reference_range(Phis[d][:, j]))

    return Phis


def shared_correlation(Xs, Ws, use_genes):
    """ Measure shared correlation strength for each CC component in each dataset. """
    n_datasets = len(Xs)
    n_cc = Ws[0].shape[1]
    C = np.zeros((n_datasets, n_cc))
    for ci in range(n_cc):
        for di, (X, W) in enumerate(zip(Xs, Ws)):
            genes = use_genes[ci]
            C[di, ci] = np.nanmean([abs(biweight_midcorrelation(X[:, g], W[:, ci])) for g in genes])
    return C


def duplicated(x):
    """ Mirror R function. """
    z = np.zeros((len(x),), dtype=bool)
    count = dict()
    for i in range(len(x)):
        z[i] = x[i] in count
        count[x[i]] = True
    return z


def align(Ws):
    """ Align data sets with dynamic time warping.
        Largest of the datasets (first in Ws) is chosen as the reference. """
    n_cc = Ws[0].shape[1]
    n_datasets = len(Ws)

    # Process each component independently
    for d in range(1, n_datasets):
        for j in range(n_cc):
            # Alignment makes sense only if the two vectors are sorted,
            # as there is no other logical correspondence between the cells in the two datasets.
            wl = Ws[0][:, j]
            ws = Ws[d][:, j]
            assert len(wl) >= len(ws)
            ars = np.argsort(ws)
            arl = np.argsort(wl)
            _, path = fastdtw(wl[arl], ws[ars])

            # Map smaller data set to larger
            P = np.array(path)
            keep = np.logical_not(duplicated(P[:, 1]))
            Pd = P[keep]

            # New values must be read from the larger vector (wl),
            # otherwise indices do not make sense
            Ws[d][ars[Pd[:, 1]], j] = Ws[0][arl[Pd[:, 0]], j]

    return Ws


class SeuratAlignmentModel:
    """
    Implementation of the Seurat multi data set alignment method,
    based on Canonical Correlation Analysis (CCA)
    and Dynamic Time Warping (DTW).
    """

    def __init__(self, n_components=20, n_metagenes=30, gene_scoring=GENE_SCORING_METHODS[0], random_state=None):
        """
        :param n_components: Number of components of final projection.
        :param n_metagenes: Number of characteristic genes for each CCA vector.
        :param gene_scoring: Correlation method to select metagenes. One of {'pearson', 'spearman'}.
        :param random_state: Random seed.
        """
        if gene_scoring not in GENE_SCORING_METHODS:
            raise ValueError("Parameter gene_scoring must be in %s." % str(GENE_SCORING_METHODS))
        self.gene_scoring = gene_scoring
        self.n_components = n_components
        self.n_metagenes = n_metagenes
        self.random_state = random_state
        self.Ws = None
        self.use_genes = None
        self.correlations = None
        self.shared_correlations = None

    def fit_transform(self, X, y):
        """
        Fit model parameters and return transformed data.
        :param X: Numeric matrix.
        :param y: Class (data set) labels.
        :return:
        """
        # Datasets are sorted by size for later alignment
        ys = sorted(set(y), key=lambda yi: sum(y == yi), reverse=True)
        index_map = dict([(yi, np.where(y == yi)[0]) for yi in ys])
        Xs = [X[index_map[yi]] for yi in ys]

        # Step 1: Run canonical correlation analysis
        if len(Xs) == 2:
            cca_model = SVDCCA(n_components=self.n_components,
                               random_state=self.random_state,
                               standardize=True)
            Ws = cca_model.fit_transform(*Xs)
            self.correlations = cca_model.correlations
        elif len(Xs) > 2:
            cca_model = MultiCCA(n_components=self.n_components,
                                 random_state=self.random_state,
                                 standardize=True)
            Ws = cca_model.fit_transform(Xs)
            self.correlations = cca_model.correlations
        else:
            raise ValueError("At least two data sets (indicated by y) are required!")

        # Step 2: Select metagenes and map to reference range
        self.use_genes = score_genes(Xs, Ws, n_metagenes=self.n_metagenes, method=self.gene_scoring)
        Phis = metagene_map(Xs, Ws, self.use_genes)

        # Step 2a: Compute shared correlations
        self.shared_correlations = shared_correlation(Xs, Ws, use_genes=self.use_genes)

        # Step 3: Run dynamic time warping
        Us = align(Phis)

        # Aggregate results according to y
        Z = np.zeros((X.shape[0], self.n_components))
        for U, yi in zip(Us, ys):
            Z[index_map[yi]] = U

        return Z

    def fit(self, X, y):
        """
        Fit model parameters.
        :param X: Numeric matrix.
        :param y: Class (data set) labels.
        :return: Ws: Model.
        """
        # Datasets are sorted by size for later alignment
        ys = sorted(set(y), key=lambda yi: sum(y == yi), reverse=True)
        index_map = dict([(yi, np.where(y == yi)[0]) for yi in ys])
        Xs = [X[index_map[yi]] for yi in ys]

        # Step 1: Run canonical correlation analysis
        if len(Xs) == 2:
            cca_model = SVDCCA(n_components=self.n_components,
                               random_state=self.random_state,
                               standardize=True)
            Ws = cca_model.fit_transform(*Xs)
            self.correlations = cca_model.correlations
        elif len(Xs) > 2:
            cca_model = MultiCCA(n_components=self.n_components,
                                 random_state=self.random_state,
                                 standardize=True)
            Ws = cca_model.fit_transform(Xs)
            self.correlations = cca_model.correlations
        else:
            raise ValueError("At least two data sets (indicated by y) are required!")

        # Step 2: Select metagenes and map to reference range
        self.use_genes = score_genes(Xs, Ws, n_metagenes=self.n_metagenes, method=self.gene_scoring)

        # Step 2a: Compute shared correlations
        self.shared_correlations = shared_correlation(Xs, Ws, use_genes=self.use_genes)

        self.Ws = Ws
        return Ws

    def transform(self, X, y, normalize=True, quantile=2.5, dtw=True, Ws=None):
        """
        Transform given data based on a model.
        :param X: Numeric matrix.
        :param y: Class (data set) labels.
        :param normalize: Boolean, whether to apply quantile normalization.
        :param quantile: float (between 0 and 49(, on which quantile will be transformed data normalized.
        :param dtw: Boolean, wheter to apply Dynamic time warping.
        :param Ws: Pretrained model.
        :return: Transformed data.
        """
        if Ws is None:
            Ws = self.Ws

        # Datasets are sorted by size for later alignment
        ys = sorted(set(y), key=lambda yi: sum(y == yi), reverse=True)
        index_map = dict([(yi, np.where(y == yi)[0]) for yi in ys])
        Xs = [X[index_map[yi]] for yi in ys]

        # Step 2: Map to reference range
        qmin = quantile
        qmax = 100 - quantile
        Us = metagene_map(Xs, Ws, self.use_genes, normalize, qmin, qmax)

        # Step 3: Run dynamic time warping
        if dtw:
            Us = align(Us)

        # Aggregate results according to y
        Z = np.zeros((X.shape[0], self.n_components))
        for U, yi in zip(Us, ys):
            Z[index_map[yi]] = U

        return Z



from collections import namedtuple
import numpy as np

from AnyQt.QtCore import Qt, QSize, QRectF, QPointF, pyqtSignal as Signal
from AnyQt.QtGui import QColor, QMouseEvent
from AnyQt.QtWidgets import QHBoxLayout, QVBoxLayout, QToolTip, \
    QGraphicsSceneMouseEvent

import pyqtgraph as pg

from Orange.data import Table
from Orange.widgets import gui, report
from Orange.widgets.settings import Setting, ContextSetting, \
    DomainContextHandler
from Orange.widgets.visualize.utils.plotutils import MouseEventDelegate
from Orange.widgets.widget import OWWidget, Input, Output, Msg

from orangecontrib.single_cell.preprocess.scpreprocess import \
    DropoutGeneSelection

ENTREZ_ID = "Entrez ID"
DropoutResults = namedtuple("DropoutResults",
                            ["zero_rate", "mean_expr", "decay",
                             "x_offset", "y_offset", "threshold"])


class States:
    WAITING, ON_CURVE, HOLDING_CURVE, MOVING_CURVE = range(4)


class DropoutGraph(pg.PlotWidget):
    pg.setConfigOption("foreground", "k")

    curve_moved = Signal(float, float)
    CURVE_PEN = pg.mkPen(color=QColor(Qt.darkCyan), width=4)
    MOVING_CURVE_PEN = pg.mkPen(color=QColor(179, 215, 255), width=4)

    def __init__(self, parent):
        super().__init__(parent, background="w")
        self.__dots = None  # type: pg.ScatterPlotItem
        self.__markers = None  # type: pg.ScatterPlotItem
        self.__curve = None  # type: pg.PlotCurveItem
        self.__decay = None  # type: float
        self.__x_offset = None  # type: float
        self.__y_offset = None  # type: float
        self._initial_x = None  # type: float
        self._initial_y = None  # type: float
        self._state = States.WAITING
        self._delegate = MouseEventDelegate(self.help_event, self.cursor_event)
        self.setMouseEnabled(False, False)
        self.hideButtons()
        self.plotItem.setContentsMargins(0, 20, 20, 0)
        self.setLabel("bottom", "Mean log2 nonzero expression")
        self.setLabel("left", "Frequency of zero expression")
        self.scene().installEventFilter(self._delegate)

    def set_data(self, results: DropoutResults, data: Table, genes: Table):
        self.__plot_dots(results.mean_expr, results.zero_rate, data)
        self.update_markers(data, genes)
        self.update_curve(results)
        self.__set_range(results.threshold, results.mean_expr)

    def update_markers(self, data: Table, genes: Table):
        self.removeItem(self.__markers)
        if data is None or self.__dots is None or genes is None:
            return
        x, y = self.__dots.getData()
        self.__plot_markers(x, y, data, genes)

    def update_curve(self, results: DropoutResults):
        self.__set_curve_params(results)
        self.__plot_curve(results)

    def __plot_dots(self, x: np.ndarray, y: np.ndarray, data: Table):
        data = list(zip(data.domain.attributes,
                        (data.X > 0).sum(axis=0),
                        np.full_like(y, len(data))))
        self.__dots = pg.ScatterPlotItem(x=x, y=y, size=5, data=data)
        self.addItem(self.__dots)

    def __plot_markers(self, x: np.ndarray, y: np.ndarray,
                       data: Table, markers: Table):
        col = markers.get_column_view(ENTREZ_ID)[0]
        mask = [str(a.attributes.get(ENTREZ_ID, None)) in col
                for a in data.domain.attributes]
        self.__markers = pg.ScatterPlotItem(
            x=x[mask], y=y[mask], size=7,
            brush=pg.mkBrush(color=QColor(Qt.magenta)))
        self.addItem(self.__markers)

    def __set_curve_params(self, results: DropoutResults):
        self.__decay = results.decay
        self.__x_offset = results.x_offset
        self.__y_offset = results.y_offset

    def __plot_curve(self, results: DropoutResults):
        self.removeItem(self.__curve)
        xmin, xmax = self.__get_xlim(results.threshold, results.mean_expr)
        x = np.arange(xmin, xmax + 0.01, 0.01)
        y = np.exp(-results.decay * (x - results.x_offset)) + results.y_offset
        pen = self.MOVING_CURVE_PEN if self._state == States.MOVING_CURVE \
            else self.CURVE_PEN
        self.__curve = pg.PlotCurveItem(
            x=x, y=y, fillLevel=1, pen=pen,
            brush=pg.mkBrush(color=QColor(0, 250, 0, 50)),
            antialias=True)
        self.addItem(self.__curve)

    def __set_range(self, threshold: float, x: np.ndarray):
        xmin, xmax = self.__get_xlim(threshold, x)
        rect = QRectF(xmin, 0, xmin + xmax, 1)
        self.setRange(rect, padding=0)

    def help_event(self, ev):
        if self.__dots is None:
            return False
        dot = self._dotAt(self.__dots.mapFromScene(ev.scenePos()))
        if dot is not None and dot.data() is not None:
            var, p, n = dot.data()
            q = round(p / n * 100, 1)
            text = f"{var.name}\nExpressed in {int(p)}/{int(n)} cells ({q}%)"
            QToolTip.showText(ev.screenPos(), text, widget=self)
            return True
        else:
            return False

    def _dotAt(self, pos: QPointF):
        pw, ph = self.__dots.pixelWidth(), self.__dots.pixelHeight()
        for s in self.__dots.points():
            sx, sy = s.pos().x(), s.pos().y()
            s2x = s2y = s.size()
            if self.__dots.opts['pxMode']:
                s2x *= pw
                s2y *= ph
            if sx + s2x > pos.x() > sx - s2x and sy + s2y > pos.y() > sy - s2y:
                return s
        return None

    def cursor_event(self, ev: QGraphicsSceneMouseEvent):
        if self.__curve is None:
            return False
        if self._state == States.HOLDING_CURVE:
            return False

        pos = self.__curve.mapFromScene(ev.scenePos())
        if self._state == States.MOVING_CURVE:
            self.curve_moved.emit(pos.x() - self._initial_x,
                                  pos.y() - self._initial_y)
            self._initial_x = pos.x()
            self._initial_y = pos.y()
            return False

        if self._on_curve(pos.x(), pos.y()):
            self._initial_x = pos.x()
            self._initial_y = pos.y()
            if self._state == States.WAITING:
                self._state = States.ON_CURVE
                self.getViewBox().setCursor(Qt.SizeAllCursor)
        else:
            self.getViewBox().setCursor(Qt.ArrowCursor)
            if self._state == States.ON_CURVE:
                self._state = States.WAITING
        return False

    def _on_curve(self, x: float, y: float):
        if self.__curve is None:
            return False
        return abs(DropoutGeneSelection.y(
            x, self.__decay, self.__x_offset, self.__y_offset) - y) < 0.01

    def mousePressEvent(self, ev: QMouseEvent):
        super().mousePressEvent(ev)
        if self._state == States.ON_CURVE:
            self._state = States.HOLDING_CURVE
            self.__curve.setPen(self.MOVING_CURVE_PEN)

    def mouseMoveEvent(self, ev: QMouseEvent):
        super().mouseMoveEvent(ev)
        if self._state == States.HOLDING_CURVE:
            self._state = States.MOVING_CURVE

    def mouseReleaseEvent(self, ev: QMouseEvent):
        super().mouseReleaseEvent(ev)
        if self._state in (States.HOLDING_CURVE, States.MOVING_CURVE):
            self._state = States.WAITING
            self.__curve.setPen(self.CURVE_PEN)

    def clear_all(self):
        self.clear()
        self.__dots = None
        self.__curve = None
        self.__decay = None
        self.__x_offset = None
        self.__y_offset = None

    @staticmethod
    def __get_xlim(threshold: float, x: np.ndarray):
        if not len(x):
            return 0, 0
        xmin = 0 if threshold == 0 else np.log2(threshold)
        return xmin, np.ceil(np.nanmax(x))


class FilterType:
    ByNumber, ByEquation = range(2)


class OWDropout(OWWidget):
    name = "Dropout Gene Selection"
    description = "Dropout-based gene selection"
    icon = 'icons/Dropout.svg'
    priority = 205

    class Inputs:
        data = Input("Data", Table, default=True)
        genes = Input("Genes", Table)

    class Outputs:
        data = Output("Data", Table)

    class Warning(OWWidget.Warning):
        less_selected = Msg("Cannot select more than {} genes.")
        missing_entrez_id = Msg("'Entred ID' is missing in Genes table.")

    settingsHandler = DomainContextHandler()
    filter_type = Setting(FilterType.ByNumber)
    n_genes = ContextSetting(1000)
    x_offset = ContextSetting(5)
    y_offset = ContextSetting(0.02)
    decay = ContextSetting(1)
    auto_commit = Setting(True)

    graph_name = "graph.plotItem"

    def __init__(self):
        super().__init__()
        self.data = None  # type: Table
        self.genes = None  # type: Table
        self.zero_rate = None  # type: np.ndarray
        self.mean_expr = None  # type: np.ndarray
        self.selected = None  # type: np.ndarray
        self.setup_gui()

    def setup_gui(self):
        self._add_graph()
        self._add_controls()

    def _add_graph(self):
        box = gui.vBox(self.mainArea, True, margin=0)
        self.graph = DropoutGraph(self)
        self.graph.curve_moved.connect(self.__manual_move)
        box.layout().addWidget(self.graph)

    def _add_controls(self):
        info_box = gui.widgetBox(self.controlArea, "Info")
        self.info_label = gui.widgetLabel(info_box, "")

        filter_box = gui.radioButtons(
            self.controlArea, self, "filter_type", box="Filter",
            orientation=QVBoxLayout(), callback=self.__filter_type_changed)

        genes_layout = QHBoxLayout()
        formula_layout = QVBoxLayout()
        genes_layout.addWidget(gui.appendRadioButton(
            filter_box, "Number of genes:", addToLayout=False))
        genes_layout.addWidget(gui.spin(
            filter_box, self, "n_genes", 0, 100000, addToLayout=False,
            callback=self.__param_changed))
        formula_layout.addWidget(gui.appendRadioButton(
            filter_box, "Apply exp(-a(x-b))+c", addToLayout=False))
        filter_box.layout().addLayout(genes_layout)
        filter_box.layout().addLayout(formula_layout)

        gui.separator(filter_box, height=1)
        coef_box = gui.hBox(filter_box, False)
        gui.separator(coef_box, width=15)
        common = dict(orientation=Qt.Horizontal,
                      callback=self.__param_changed,
                      alignment=Qt.AlignRight, controlWidth=60)
        gui.doubleSpin(
            coef_box, self, "decay", 0.0, 10.0, 0.01, label="a:", **common)
        gui.doubleSpin(
            coef_box, self, "x_offset", 0.0, 10.0, 0.01, label="b:", **common)
        gui.doubleSpin(
            coef_box, self, "y_offset", 0.0, 1.0, 0.01, label="c:", **common)

        gui.rubber(self.controlArea)
        gui.auto_commit(self.controlArea, self, "auto_commit",
                        "Send Selection", "Send Automatically")

        self.setup_info_label()
        self.enable_controls()

    def __filter_type_changed(self):
        self.enable_controls()
        self.__param_changed()

    def __param_changed(self):
        self.update_selection()
        self.setup_info_label()
        self.commit()

    def __manual_move(self, delta_x, delta_y):
        self.x_offset += delta_x
        self.y_offset += delta_y
        if self.x_offset < 0:
            self.x_offset = 0
        if self.y_offset < 0:
            self.y_offset = 0
        self.filter_type = FilterType.ByEquation
        self.__filter_type_changed()

    @property
    def filter_by_nr_of_genes(self):
        return self.filter_type == FilterType.ByNumber

    @Inputs.data
    def set_data(self, data):
        self.closeContext()
        self.data = data
        self.openContext(data)
        self.select_genes()
        self.setup_info_label()
        self.unconditional_commit()

    @Inputs.genes
    def set_genes(self, genes):
        self.genes = genes
        self.check_genes()
        self.graph.update_markers(self.data, self.genes)

    def check_genes(self):
        self.Warning.missing_entrez_id.clear()
        if self.genes:
            if ENTREZ_ID not in self.genes.domain:
                self.Warning.missing_entrez_id()
                self.genes = None

    def select_genes(self):
        self.selected = None
        self.graph.clear_all()
        self.Warning.less_selected.clear()
        if not self.data:
            return

        selector = self.__get_selector()
        self.zero_rate, self.mean_expr = selector.detection(self.data.X)
        results = self.__select(selector)
        self.graph.set_data(DropoutResults(*results), self.data, self.genes)

    def __get_selector(self):
        kwargs = {"decay": self.decay, "y_offset": self.y_offset}
        if self.filter_by_nr_of_genes:
            kwargs["n_genes"] = self.n_genes
        else:
            kwargs["x_offset"] = self.x_offset
        return DropoutGeneSelection(**kwargs)

    def __select(self, selector):
        self.selected = selector.select_genes(self.zero_rate, self.mean_expr)
        n_selected = sum(self.selected)
        if n_selected < self.n_genes and self.filter_by_nr_of_genes:
            self.Warning.less_selected(n_selected)
        self.n_genes = n_selected
        self.decay = selector.decay
        self.x_offset = selector.x_offset
        self.y_offset = selector.y_offset
        return (self.zero_rate, self.mean_expr, selector.decay,
                selector.x_offset, selector.y_offset, selector.threshold)

    def update_selection(self):
        self.Warning.less_selected.clear()
        if not self.data:
            return
        selector = self.__get_selector()
        self.graph.update_curve(DropoutResults(*self.__select(selector)))

    def setup_info_label(self):
        text = "No data on input."
        if self.selected is not None:
            k = sum(self.selected)
            n, m = len(self.data), len(self.data.domain.attributes)
            ks = "s" if k != 1 else ""
            ns, ms = "s" if n != 1 else "", "s" if m != 1 else ""
            text = f"Data with {n} cell{ns} and {m} gene{ms}" \
                   f"\n{k} gene{ks} in selection"
        self.info_label.setText(text)

    def enable_controls(self):
        self.controls.n_genes.setEnabled(self.filter_by_nr_of_genes)
        self.controls.decay.setEnabled(not self.filter_by_nr_of_genes)
        self.controls.x_offset.setEnabled(not self.filter_by_nr_of_genes)
        self.controls.y_offset.setEnabled(not self.filter_by_nr_of_genes)

    def commit(self):
        data = None
        if self.selected is not None:
            data = DropoutGeneSelection.filter_columns(self.data,
                                                       self.selected)
        self.Outputs.data.send(data)

    def sizeHint(self):
        return super().sizeHint().expandedTo(QSize(800, 400))

    def send_report(self):
        if self.selected is None:
            return
        self.report_plot()
        self.report_caption(report.render_items_vert((
            ("Number of genes", self.n_genes),
            ("a", round(self.decay, 2)),
            ("b", round(self.x_offset, 2)),
            ("a", round(self.y_offset, 2)))))


if __name__ == "__main__":
    from Orange.widgets.utils.widgetpreview import WidgetPreview

    table = Table("https://datasets.orange.biolab.si/sc/aml-1k.tab.gz")
    WidgetPreview(OWDropout).run(set_data=table)



from AnyQt.QtWidgets import QFormLayout
from AnyQt.QtGui import QColor
from AnyQt.QtCore import Qt

import numpy as np
from collections import OrderedDict
import pyqtgraph as pg
from pyqtgraph.graphicsItems import LegendItem
from pyqtgraph import functions as fn
from scipy.stats import multivariate_normal as mvn

from Orange.widgets.settings import Setting, ContextSetting, PerfectDomainContextHandler
from Orange.data import Table, Domain, ContinuousVariable, DiscreteVariable
from Orange.widgets import widget, gui
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.widgets.widget import Input, Output
from Orange.widgets.utils.sql import check_sql_input
from Orange.widgets.utils.annotated_data import add_columns

from orangecontrib.single_cell.preprocess.alignment import SeuratAlignmentModel

MAX_COMPONENTS = 50
MAX_COMPONENTS_DEFAULT = 50
MAX_GENES = 100

SCORINGS = OrderedDict([
    ("Pearson correlation", "pearson"),
    ("Spearman correlation", "spearman"),
    ("Biweights midcorrelation", "bicor"),
])


class MyLegendItem(LegendItem.ItemSample):
    def paint(self, p, *args):
        opts = self.item.opts
        opts['pen'].setWidth(2)
        p.setPen(fn.mkPen(opts['pen']))
        p.drawLine(0, 10, 9, 10)


def smooth_correlations(M, offset=2):
    """
    Smoothed correlations; Convolve with a small Gaussian bump.
    :param M: model.shared_correlations
    :param offset: First few values to leave unchanged
    :return:
    """
    g = mvn.pdf(x=np.linspace(-offset, offset, 2 * offset + 1))
    Z = np.zeros(M.shape)
    for i, m in enumerate(M):
        Z[i, :offset] = m[:offset]
        Z[i, offset:] = np.convolve(m, g, mode="same")[offset:]
    return Z


def interpolate_nans(A):
    ok = ~np.isnan(A)
    xp = ok.ravel().nonzero()[0]
    fp = A[ok]
    x = np.isnan(A).ravel().nonzero()[0]
    if len(xp) > 0:
        A[np.isnan(A)] = np.interp(x, xp, fp)
    return A


class OWAlignDatasets(widget.OWWidget):
    name = "Align Datasets"
    description = "Alignment of multiple datasets with a diagram of correlation visualization."
    icon = "icons/AlignDatasets.svg"
    priority = 240

    class Inputs:
        data = Input("Data", Table)

    class Outputs:
        transformed_data = Output("Transformed Data", Table)
        genes_components = Output("Genes per n. Components", Table)

    settingsHandler = PerfectDomainContextHandler()
    axis_labels = ContextSetting(10)
    source_id = ContextSetting(None)
    ncomponents = ContextSetting(20)
    ngenes = ContextSetting(30)
    scoring = ContextSetting(list(SCORINGS.keys())[0])
    quantile_normalization = ContextSetting(False)
    quantile_normalization_perc = ContextSetting(2.5)
    dynamic_time_warping = ContextSetting(False)

    auto_update = Setting(True)
    auto_commit = Setting(True)

    graph_name = "plot.plotItem"

    class Error(widget.OWWidget.Error):
        no_features = widget.Msg("At least 1 feature is required")
        no_instances = widget.Msg("At least 2 data instances are required for each class")
        no_class = widget.Msg("At least 1 Discrete class variable is required")
        nan_class = widget.Msg(
            "Data contains undefined instances for the selected Data source indicator")
        nan_input = widget.Msg("Input data contains non numeric values")
        sparse_data = widget.Msg("Sparse data is not supported")
        only_one_dataset = widget.Msg("Data source indicator attribute column must indicate at least two datasets.")

    def __init__(self):
        super().__init__()
        self.data = None
        self.source_id = None
        self._mas = None
        self._Ws = None
        self._transformed = None
        self._components = None
        self._use_genes = None
        self._shared_correlations = None
        self._transformed_table = None
        self._line = False
        self._feature_model = DomainModel(valid_types=DiscreteVariable, separators=False)
        self._feature_model.set_domain(None)
        self._init_mas()
        self._legend = None
        form = QFormLayout(
            labelAlignment=Qt.AlignLeft,
            formAlignment=Qt.AlignLeft,
            fieldGrowthPolicy=QFormLayout.AllNonFixedFieldsGrow,
            verticalSpacing=10
        )
        # Data source indicator
        box = gui.vBox(self.controlArea, "Data source indicator")

        gui.comboBox(
            box, self, "source_id", sendSelectedValue=True,
            callback=self._update_combo_source_id,
            model=self._feature_model,
        )

        # Canonical correlation analysis
        box = gui.vBox(self.controlArea, "Canonical correlation analysis")
        gui.spin(
            box, self, "ncomponents", 1, MAX_COMPONENTS,
            callback=self._update_selection_component_spin,
            keyboardTracking=False,
            label="Num. of components"
        )

        # Shared genes
        box = gui.vBox(self.controlArea, "Shared genes")
        gui.spin(
            box, self, "ngenes", 1, MAX_GENES,
            callback=self._update_ngenes_spin,
            keyboardTracking=False,
        )
        form.addRow(
            "Num. of genes",
            self.controls.ngenes
        )

        gui.comboBox(
            box, self, "scoring",
            callback=self._update_scoring_combo,
            items=list(SCORINGS.keys()), sendSelectedValue=True,
            editable=False,
        )
        form.addRow(
            "Scoring:",
            self.controls.scoring
        )

        box.layout().addLayout(form)

        # Post-processing
        box = gui.vBox(self.controlArea, "Post-processing")
        gui.doubleSpin(
            box, self, "quantile_normalization_perc", minv=0, maxv=49, step=5e-1,
            callback=self._update_quantile_normalization,
            checkCallback=self._update_quantile_normalization,
            controlWidth=80, alignment=Qt.AlignRight,
            label="Quantile normalization", checked="quantile_normalization",
        )
        self.controls.quantile_normalization_perc.setSuffix("%")
        b = gui.vBox(box)
        gui.checkBox(
            b, self, "dynamic_time_warping",
            callback=self._update_dynamic_time_warping,
            label="Dynamic time warping"
        )

        self.controlArea.layout().addStretch()

        gui.auto_commit(self.controlArea, self, "auto_commit", "Apply",
                        callback=self._invalidate_selection(),
                        checkbox_label="Apply automatically")

        self.plot = pg.PlotWidget(background="w")

        axis = self.plot.getAxis("bottom")
        axis.setLabel("Correlation components")
        axis = self.plot.getAxis("left")
        axis.setLabel("Correlation strength")
        self.plot_horlabels = []
        self.plot_horlines = []

        self.plot.getViewBox().setMenuEnabled(False)
        self.plot.getViewBox().setMouseEnabled(False, False)
        self.plot.showGrid(True, True, alpha=0.5)
        self.plot.setRange(xRange=(0.0, 1.0), yRange=(0.0, 1.0))

        self.mainArea.layout().addWidget(self.plot)

    @Inputs.data
    @check_sql_input
    def set_data(self, data):
        self.closeContext()
        self.clear_messages()
        self.clear()
        self.information()
        self.clear_outputs()
        self._feature_model.set_domain(None)
        self.data = data

        if self.data:
            self._feature_model.set_domain(self.data.domain)
            if self._feature_model:
                # if source id is available we assume that it is the feature that describes a dataset source
                if "Source ID" in self.data.domain:
                    self.source_id = self.data.domain["Source ID"]
                self.openContext(self.data.domain)
                if self.source_id is None or self.source_id == '':
                    for model in self._feature_model:
                        y = np.array(self.data.get_column_view(model)[0], dtype=np.float64)
                        _, counts = np.unique(y, return_counts=True)
                        if np.isfinite(y).all() and min(counts) > 1:
                            self.source_id = model
                            self._reset_max_components()
                            break

                if not self.source_id:
                    self.Error.nan_class()
                    return
                if len(self.data.domain.attributes) == 0:
                    self.Error.no_features()
                    return
                if len(self.data) == 0:
                    self.Error.no_instances()
                    return
                if np.isnan(self.data.X).any():
                    self.Error.nan_input()
                    return
                y = np.array(self.data.get_column_view(self.source_id)[0], dtype=np.float64)
                _, counts = np.unique(y, return_counts=True)
                if min(counts) < 2:
                    self.Error.no_instances()
                    return
                self._reset_max_components()
                self.fit()

            else:
                self.Error.no_class()
                self.clear()
                return

    def fit(self):
        if self.data is None:
            return
        global MAX_COMPONENTS
        if self.ncomponents > MAX_COMPONENTS:
            self.ncomponents = MAX_COMPONENTS

        X = self.data.X
        y = self.data.get_column_view(self.source_id)[0]

        if len(set(y)) < 2:
            self.Error.only_one_dataset()
            return
        self._init_mas()

        self._Ws = self._mas.fit(X, y)
        self._shared_correlations = self._mas.shared_correlations
        if np.isnan(np.sum(self._shared_correlations)):
            self._shared_correlations = np.array([interpolate_nans(x) for x in self._shared_correlations])
        self._use_genes = self._mas.use_genes

        self._setup_plot()
        if self.auto_commit:
            self.commit()

    def clear(self):
        self.data = None
        self.source_id = None
        self._mas = None
        self._Ws = None
        self._transformed = None
        self._transformed_table = None
        self._components = None
        self._use_genes = None
        self._shared_correlations = None
        self._feature_model.set_domain(None)
        self.clear_plot()

    def clear_legend(self):
        if self._legend is None:
            return

        scene = self._legend.scene()
        if scene is None:
            return

        scene.removeItem(self._legend)
        self._legend = None

    def clear_plot(self):
        self.clear_legend()
        self._line = False
        self.plot_horlabels = []
        self.plot_horlines = []
        self._mas = None
        self._setup_plot()

    def clear_outputs(self):
        self.Outputs.transformed_data.send(None)
        self.Outputs.genes_components.send(None)

    def _reset_max_components(self):
        y = np.array(self.data.get_column_view(self.source_id)[0], dtype=np.float64)
        _, counts = np.unique(y, return_counts=True)
        global MAX_COMPONENTS
        if min(counts) < MAX_COMPONENTS_DEFAULT or len(
                self.data.domain.attributes) < MAX_COMPONENTS_DEFAULT:
            MAX_COMPONENTS = min(min(counts), len(self.data.domain.attributes)) - 1
            if self.ncomponents > MAX_COMPONENTS:
                self.ncomponents = MAX_COMPONENTS // 2
            self.controls.ncomponents.setMaximum(MAX_COMPONENTS)
        else:
            MAX_COMPONENTS = MAX_COMPONENTS_DEFAULT
            self.ncomponents = 20
            self.controls.ncomponents.setMaximum(MAX_COMPONENTS)

    def _init_mas(self):
        self._mas = SeuratAlignmentModel(
            n_components=MAX_COMPONENTS,
            n_metagenes=self.ngenes,
            gene_scoring=SCORINGS[self.scoring],
        )

    def get_model(self):
        if self.data is None:
            return

        self.fit()
        self._setup_plot()
        self.commit()

    def _setup_plot(self):
        self.plot.clear()
        if self._mas is None:
            return

        shared_correlations = self._shared_correlations
        p = MAX_COMPONENTS

        # Colors chosen based on: http://colorbrewer2.org/?type=qualitative&scheme=Set1&n=9
        colors = ['#e41a1c', '#377eb8', '#4daf4a', '#984ea3', '#ff7f00', '#ffff33', '#a65628',
                  '#f781bf', '#999999']

        self.clear_legend()
        self._legend = self.plot.addLegend(offset=(-1, 1))
        # correlation lines
        offset = 2
        if MAX_COMPONENTS > 2 * offset + 1:
            smoothed_correlations = smooth_correlations(shared_correlations, offset=offset)
        else:
            smoothed_correlations = shared_correlations
        plotitem = dict()
        for i, corr in enumerate(smoothed_correlations):
            plotitem[i] = self.plot.plot(np.arange(p), corr,
                                         pen=pg.mkPen(QColor(colors[i]), width=2),
                                         antialias=True)  # name=self.source_id.values[i]
        # self.plot.plotItem.legend.addItem(3, "maximum value")

        for i in range(len(plotitem)):
            self._legend.addItem(MyLegendItem(pg.ScatterPlotItem(pen=colors[i])),
                                 self.source_id.values[i])

        # vertical movable line
        cutpos = self.ncomponents - 1
        self._line = pg.InfiniteLine(
            angle=90, pos=cutpos, movable=True, bounds=(0, p - 1))
        self._line.setCursor(Qt.SizeHorCursor)
        self._line.setPen(pg.mkPen(QColor(Qt.black), width=2))
        self._line.sigPositionChanged.connect(self._on_cut_changed)
        self.plot.addItem(self._line)

        # horizontal lines
        self.plot_horlines = tuple(
            pg.PlotCurveItem(pen=pg.mkPen(QColor(colors[i]), style=Qt.DashLine)) for i in
            range(len(shared_correlations))
        )
        self.plot_horlabels = tuple(
            pg.TextItem(color=QColor('k'), anchor=(0, 1)) for _ in range(len(shared_correlations))
        )

        for item in self.plot_horlabels + self.plot_horlines:
            self.plot.addItem(item)
        self._set_horline_pos()

        # self.plot.setRange(xRange=(0.0, p - 1), yRange=(0.0, 1.0))
        self.plot.setXRange(0.0, p - 1, padding=0)
        self.plot.setYRange(0.0, 1.0, padding=0)
        self._update_axis()

    def _set_horline_pos(self):
        cutidx = self.ncomponents - 1
        for line, label, curve in zip(self.plot_horlines, self.plot_horlabels,
                                      self._shared_correlations):
            y = curve[cutidx]
            line.setData([-1, cutidx], 2 * [y])
            label.setPos(cutidx, y)
            label.setPlainText("{:.3f}".format(y))

    def _on_cut_changed(self, line):
        # cut changed by means of a cut line over the scree plot.
        value = int(round(line.value()))
        components = value + 1

        if not (self.ncomponents == 0 and
                components == len(self._components)):
            self.ncomponents = components

        self._line.setValue(value)
        self._set_horline_pos()
        self.commit()

    def _update_selection_component_spin(self):
        # cut changed by "ncomponents" spin.
        if self._mas is None:
            self._invalidate_selection()
            return

        if np.floor(self._line.value()) + 1 != self.ncomponents:
            self._line.setValue(self.ncomponents - 1)

        self.commit()

    def _invalidate_selection(self):
        if self.data is not None:
            self._transformed = None
            self.commit()

    def _update_scoring_combo(self):
        self.fit()
        self._invalidate_selection()

    def _update_dynamic_time_warping(self):
        self._invalidate_selection()

    def _update_quantile_normalization(self):
        self._invalidate_selection()

    def _update_ngenes_spin(self):
        self.clear_plot()
        if self.data is None:
            return
        if self._has_nan_classes():
            self.Error.nan_class()
            return
        self.clear_messages()
        self.fit()
        self._invalidate_selection()

    def _update_combo_source_id(self):
        self.clear_plot()
        if self.data is None:
            return
        y = np.array(self.data.get_column_view(self.source_id)[0], dtype=np.float64)
        _, counts = np.unique(y, return_counts=True)
        if min(counts) < 2:
            self.Error.no_instances()
            return
        self._reset_max_components()
        if self._has_nan_classes():
            self.Error.nan_class()
            return
        self.clear_messages()
        self.fit()
        self._invalidate_selection()

    def _update_axis(self):
        p = MAX_COMPONENTS
        axis = self.plot.getAxis("bottom")
        d = max((p - 1) // (self.axis_labels - 1), 1)
        axis.setTicks([[(i, str(i + 1)) for i in range(0, p, d)]])

    def _has_nan_classes(self):
        y = np.array(self.data.get_column_view(self.source_id)[0], dtype=np.float64)
        return not np.isfinite(y).all()

    def commit(self):
        transformed_table = meta_genes = None
        if self._mas is not None:
            # Compute the full transform (MAX_COMPONENTS components) only once.
            if self._transformed is None:
                X = self.data.X
                y = self.data.get_column_view(self.source_id)[0]
                self._transformed = self._mas.transform(X, y, normalize=self.quantile_normalization,
                                                        quantile=self.quantile_normalization_perc,
                                                        dtw=self.dynamic_time_warping)

                attributes = tuple(ContinuousVariable.make("CCA{}".format(x + 1)) for x in
                                   range(MAX_COMPONENTS))
                dom = Domain(
                    attributes,
                    self.data.domain.class_vars,
                    self.data.domain.metas
                )

                # Meta-genes
                meta_genes = self.data.transform(dom)
                genes_components = np.zeros((self.data.X.shape[1], MAX_COMPONENTS))
                for key, genes in self._mas.use_genes.items():
                    for gene in genes:
                        genes_components[gene - 1, key] = genes.index(gene) + 1
                genes_components[genes_components == 0] = np.NaN
                meta_genes.X = genes_components
                self.meta_genes = Table.from_numpy(Domain(attributes), genes_components)

                # Transformed data
                transformed = self._transformed
                new_domain = add_columns(self.data.domain, attributes=attributes)
                transformed_table_temp = self.data.transform(new_domain)
                transformed_table_temp.X[:, -MAX_COMPONENTS:] = transformed
                self.transformed_table = Table.from_table(dom, transformed_table_temp)

            ncomponents_attributes = tuple(ContinuousVariable.make("CCA{}".format(x + 1)) for x in
                                           range(self.ncomponents))
            ncomponents_domain = Domain(
                ncomponents_attributes,
                self.data.domain.class_vars,
                self.data.domain.metas
            )

            meta_genes = self.meta_genes.transform(Domain(ncomponents_attributes))
            transformed_table = self.transformed_table.transform(ncomponents_domain)

        self.Outputs.transformed_data.send(transformed_table)
        self.Outputs.genes_components.send(meta_genes)

    def send_report(self):
        if self.data is None:
            return
        self.report_items((
            ("Source ID", self.source_id),
            ("Selected num. of components", self.ncomponents),
            ("Selected num. of genes", self.ngenes),
            ("Scoring", self.scoring),
            ("Quantile normalization", True if self.quantile_normalization else "False"),
            ("Quantile normalization percentage",
             self.quantile_normalization_perc if self.quantile_normalization else False),
            ("Dynamic time warping", True if self.dynamic_time_warping else "False")
        ))
        self.report_plot()

    """
    @classmethod
    def migrate_settings(cls, settings, version):
        if "variance_covered" in settings:
            # Due to the error in gh-1896 the variance_covered was persisted
            # as a NaN value, causing a TypeError in the widgets `__init__`.
            vc = settings["variance_covered"]
            if isinstance(vc, numbers.Real):
                if np.isfinite(vc):
                    vc = int(vc)
                else:
                    vc = 100
                settings["variance_covered"] = vc
        if settings.get("ncomponents", 0) > MAX_COMPONENTS:
            settings["ncomponents"] = MAX_COMPONENTS
    """


def main():
    import gc
    import Orange
    from AnyQt.QtWidgets import QApplication
    app = QApplication([])
    w = OWAlignDatasets()
    in_file = Orange.data.Table("iris")
    data = Table(in_file)
    w.set_data(data)
    w.show()
    w.raise_()
    rval = w.exec()
    w.deleteLater()
    del w
    app.processEvents()
    gc.collect()
    return rval


if __name__ == "__main__":
    main()
